"""Utilities for with-statement contexts.  See PEP 343.

Original source code: https://hg.python.org/cpython/file/3.4/Lib/contextlib.py

Not implemented:
 - redirect_stdout;

"""

import sys
from collections import deque
from ucontextlib import *


class closing(object):
    """Context to automatically close something at the end of a block.

    Code like this:

        with closing(<module>.open(<arguments>)) as f:
            <block>

    is equivalent to this:

        f = <module>.open(<arguments>)
        try:
            <block>
        finally:
            f.close()

    """

    def __init__(self, thing):
        self.thing = thing

    def __enter__(self):
        return self.thing

    def __exit__(self, *exc_info):
        self.thing.close()


class suppress:
    """Context manager to suppress specified exceptions

    After the exception is suppressed, execution proceeds with the next
    statement following the with statement.

         with suppress(FileNotFoundError):
             os.remove(somefile)
         # Execution still resumes here if the file was already removed
    """

    def __init__(self, *exceptions):
        self._exceptions = exceptions

    def __enter__(self):
        pass

    def __exit__(self, exctype, excinst, exctb):
        # Unlike isinstance and issubclass, CPython exception handling
        # currently only looks at the concrete type hierarchy (ignoring
        # the instance and subclass checking hooks). While Guido considers
        # that a bug rather than a feature, it's a fairly hard one to fix
        # due to various internal implementation details. suppress provides
        # the simpler issubclass based semantics, rather than trying to
        # exactly reproduce the limitations of the CPython interpreter.
        #
        # See http://bugs.python.org/issue12029 for more details
        return exctype is not None and issubclass(exctype, self._exceptions)


# Inspired by discussions on http://bugs.python.org/issue13585
class ExitStack(object):
    """Context manager for dynamic management of a stack of exit callbacks

    For example:

        with ExitStack() as stack:
            files = [stack.enter_context(open(fname)) for fname in filenames]
            # All opened files will automatically be closed at the end of
            # the with statement, even if attempts to open files later
            # in the list raise an exception

    """

    def __init__(self):
        self._exit_callbacks = []

    def pop_all(self):
        """Preserve the context stack by transferring it to a new instance"""
        new_stack = type(self)()
        new_stack._exit_callbacks = self._exit_callbacks
        self._exit_callbacks = []
        return new_stack

    def _push_cm_exit(self, cm, cm_exit):
        """Helper to correctly register callbacks to __exit__ methods"""

        def _exit_wrapper(*exc_details):
            return cm_exit(cm, *exc_details)

        self.push(_exit_wrapper)

    def push(self, exit):
        """Registers a callback with the standard __exit__ method signature

        Can suppress exceptions the same way __exit__ methods can.

        Also accepts any object with an __exit__ method (registering a call
        to the method instead of the object itself)
        """
        # We use an unbound method rather than a bound method to follow
        # the standard lookup behaviour for special methods
        _cb_type = type(exit)
        try:
            exit_method = _cb_type.__exit__
        except AttributeError:
            # Not a context manager, so assume its a callable
            self._exit_callbacks.append(exit)
        else:
            self._push_cm_exit(exit, exit_method)
        return exit  # Allow use as a decorator

    def callback(self, callback, *args, **kwds):
        """Registers an arbitrary callback and arguments.

        Cannot suppress exceptions.
        """

        def _exit_wrapper(exc_type, exc, tb):
            callback(*args, **kwds)

        self.push(_exit_wrapper)
        return callback  # Allow use as a decorator

    def enter_context(self, cm):
        """Enters the supplied context manager

        If successful, also pushes its __exit__ method as a callback and
        returns the result of the __enter__ method.
        """
        # We look up the special methods on the type to match the with statement
        _cm_type = type(cm)
        _exit = _cm_type.__exit__
        result = _cm_type.__enter__(cm)
        self._push_cm_exit(cm, _exit)
        return result

    def close(self):
        """Immediately unwind the context stack"""
        self.__exit__(None, None, None)

    def __enter__(self):
        return self

    def __exit__(self, *exc_details):
        received_exc = exc_details[0] is not None
        # Callbacks are invoked in LIFO order to match the behaviour of
        # nested context managers
        suppressed_exc = False
        pending_raise = False
        while self._exit_callbacks:
            cb = self._exit_callbacks.pop()
            try:
                if cb(*exc_details):
                    suppressed_exc = True
                    pending_raise = False
                    exc_details = (None, None, None)
            except:
                exc_details = sys.exc_info()
                pending_raise = True
        if pending_raise:
            raise exc_details[1]
        return received_exc and suppressed_exc



try:
    from importlib.metadata import version, PackageNotFoundError

    try:
        __version__ = version("mpremote")
    except PackageNotFoundError:
        # Error loading package version (e.g. running from source).
        __version__ = "0.0.0-local"
except ImportError:
    # importlib.metadata not available (e.g. CPython <3.8 without
    # importlib_metadata compatibility package installed).
    __version__ = "0.0.0-unknown"



"""
The MIT License (MIT)

Copyright (c) 2013-2022 Ibrahim Abdelkader <iabdalkader@openmv.io>
Copyright (c) 2013-2022 Kwabena W. Agyeman <kwagyeman@openmv.io>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

HTS221 driver driver for MicroPython.
Original source: https://github.com/ControlEverythingCommunity/HTS221/blob/master/Python/HTS221.py

Example usage:

import time
import hts221
from machine import Pin, I2C

bus = I2C(1, scl=Pin(15), sda=Pin(14))
hts = hts221.HTS221(bus)

while (True):
    rH   = hts.humidity()
    temp = hts.temperature()
    print ("rH: %.2f%% T: %.2fC" %(rH, temp))
    time.sleep_ms(100)
"""

import struct


class HTS221:
    def __init__(self, i2c, data_rate=1, address=0x5F):
        self.bus = i2c
        self.odr = data_rate
        self.slv_addr = address

        # Set configuration register
        # Humidity and temperature average configuration
        self.bus.writeto_mem(self.slv_addr, 0x10, b"\x1B")

        # Set control register
        # PD | BDU | ODR
        cfg = 0x80 | 0x04 | (self.odr & 0x3)
        self.bus.writeto_mem(self.slv_addr, 0x20, bytes([cfg]))

        # Read Calibration values from non-volatile memory of the device
        # Humidity Calibration values
        self.H0 = self._read_reg(0x30, 1) / 2
        self.H1 = self._read_reg(0x31, 1) / 2
        self.H2 = self._read_reg(0x36, 2)
        self.H3 = self._read_reg(0x3A, 2)

        # Temperature Calibration values
        raw = self._read_reg(0x35, 1)
        self.T0 = ((raw & 0x03) * 256) + self._read_reg(0x32, 1)
        self.T1 = ((raw & 0x0C) * 64) + self._read_reg(0x33, 1)
        self.T2 = self._read_reg(0x3C, 2)
        self.T3 = self._read_reg(0x3E, 2)

    def _read_reg(self, reg_addr, size):
        fmt = "B" if size == 1 else "H"
        reg_addr = reg_addr if size == 1 else reg_addr | 0x80
        return struct.unpack(fmt, self.bus.readfrom_mem(self.slv_addr, reg_addr, size))[0]

    def humidity(self):
        rH = self._read_reg(0x28, 2)
        return (self.H1 - self.H0) * (rH - self.H2) / (self.H3 - self.H2) + self.H0

    def temperature(self):
        temp = self._read_reg(0x2A, 2)
        if temp > 32767:
            temp -= 65536
        return ((self.T1 - self.T0) / 8.0) * (temp - self.T2) / (self.T3 - self.T2) + (
            self.T0 / 8.0
        )



try:
    # Check if deflate is available.
    import deflate
    import io
except ImportError:
    print("SKIP")
    raise SystemExit

# Check if compression is enabled.
if not hasattr(deflate.DeflateIO, "write"):
    print("SKIP")
    raise SystemExit

# Simple compression & decompression.
b = io.BytesIO()
g = deflate.DeflateIO(b, deflate.RAW)
data = b"micropython"
N = 10
for i in range(N):
    g.write(data)
g.close()
result_raw = b.getvalue()
print(len(result_raw) < len(data) * N)
b = io.BytesIO(result_raw)
g = deflate.DeflateIO(b, deflate.RAW)
print(g.read())

# Same, but using a context manager.
b = io.BytesIO()
with deflate.DeflateIO(b, deflate.RAW) as g:
    for i in range(N):
        g.write(data)
result_raw = b.getvalue()
print(len(result_raw) < len(data) * N)
b = io.BytesIO(result_raw)
with deflate.DeflateIO(b, deflate.RAW) as g:
    print(g.read())

# Writing to a closed underlying stream.
b = io.BytesIO()
g = deflate.DeflateIO(b, deflate.RAW)
g.write(b"micropython")
b.close()
try:
    g.write(b"micropython")
except ValueError:
    print("ValueError")

# Writing to a closed DeflateIO.
b = io.BytesIO()
g = deflate.DeflateIO(b, deflate.RAW)
g.write(b"micropython")
g.close()
try:
    g.write(b"micropython")
except OSError:
    print("OSError")


def decompress(data, *args):
    buf = io.BytesIO(data)
    with deflate.DeflateIO(buf, *args) as g:
        return g.read()


def compress(data, *args):
    b = io.BytesIO()
    with deflate.DeflateIO(b, *args) as g:
        g.write(data)
    return b.getvalue()


def compress_error(data, *args):
    try:
        compress(data, *args)
    except OSError:
        print("OSError")
    except ValueError:
        print("ValueError")


# More test patterns.
PATTERNS_RAW = (
    (b"0", b"3\x00\x00"),
    (b"a", b"K\x04\x00"),
    (b"0" * 100, b"3\xa0\x03\x00\x00"),
    (
        bytes(range(64)),
        b"c`dbfaec\xe7\xe0\xe4\xe2\xe6\xe1\xe5\xe3\x17\x10\x14\x12\x16\x11\x15\x13\x97\x90\x94\x92\x96\x91\x95\x93WPTRVQUS\xd7\xd0\xd4\xd2\xd6\xd1\xd5\xd370426153\xb7\xb0\xb4\xb2\xb6\xb1\xb5\xb3\x07\x00",
    ),
)
for unpacked, packed in PATTERNS_RAW:
    print(compress(unpacked) == packed)
    print(compress(unpacked, deflate.RAW) == packed)

# Verify header and checksum format.
unpacked = b"hello"
packed = b"\xcbH\xcd\xc9\xc9\x07\x00"


def check_header(n, a, b):
    if a == b:
        print(n)
    else:
        print(n, a, b)


check_header("RAW", compress(unpacked, deflate.RAW), packed)
check_header(
    "ZLIB(9)", compress(unpacked, deflate.ZLIB, 9), b"\x18\x95" + packed + b"\x06,\x02\x15"
)
check_header(
    "ZLIB(15)", compress(unpacked, deflate.ZLIB, 15), b"\x78\x9c" + packed + b"\x06,\x02\x15"
)
check_header(
    "GZIP",
    compress(unpacked, deflate.GZIP, 9),
    b"\x1f\x8b\x08\x00\x00\x00\x00\x00\x04\x03" + packed + b"\x86\xa6\x106\x05\x00\x00\x00",
)

# Valid wbits values.
compress_error(unpacked, deflate.RAW, -1)
print(len(compress(unpacked, deflate.RAW, 0)))
compress_error(unpacked, deflate.RAW, 1)
compress_error(unpacked, deflate.RAW, 4)
for i in range(5, 16):
    print(len(compress(unpacked, deflate.RAW, i)))
compress_error(unpacked, deflate.RAW, 16)

# Invalid values for format.
compress_error(unpacked, -1)
compress_error(unpacked, 5)

# Fill buf with a predictable pseudorandom sequence.
buf = bytearray(1024)
lfsr = 1 << 15 | 1
for i in range(len(buf)):
    bit = (lfsr ^ (lfsr >> 1) ^ (lfsr >> 3) ^ (lfsr >> 12)) & 1
    lfsr = (lfsr >> 1) | (bit << 15)
    buf[i] = lfsr & 0xFF

# Verify that compression improves as the window size increases.
prev_len = len(buf)
for wbits in range(5, 10):
    result = compress(buf, deflate.RAW, wbits)
    next_len = len(result)
    print(next_len < prev_len and decompress(result, deflate.RAW, wbits) == buf)
    prev_len = next_len

# Verify that compression is optimal: in the bytes below, the final "123" should be
# compressed by referencing the "123" just before it, and not the one all the way back
# at the start of the bytes.
compressed = compress(b"1234567890abcdefghijklmnopqrstuvwxyz123123", deflate.RAW)
print(len(compressed), compressed)



import sys
import unittest


class TestModuleImport(unittest.TestCase):
    def test_ModuleImportPath(self):
        try:
            from sub.sub import imported
            assert imported
        except ImportError:
            print("This test is intended to be run with unittest discover"
                  "from the unittest-discover/tests dir. sys.path:", sys.path)
            raise



class A:
    def __bool__(self):
        print('__bool__')
        return True
    def __len__(self):
        print('__len__')
        return 1

class B:
    def __len__(self):
        print('__len__')
        return 0

print(bool(A()))
print(len(A()))
print(bool(B()))
print(len(B()))



# Test notification-specific behavior.

import sys

# ruff: noqa: E402
sys.path.append("")

from micropython import const
import machine
import time

import asyncio
import aioble
import bluetooth

TIMEOUT_MS = 5000

SERVICE_UUID = bluetooth.UUID("A5A5A5A5-FFFF-9999-1111-5A5A5A5A5A5A")
CHAR_UUID = bluetooth.UUID("00000000-1111-2222-3333-444444444444")


# Acting in peripheral role.
async def instance0_task():
    service = aioble.Service(SERVICE_UUID)
    characteristic = aioble.Characteristic(service, CHAR_UUID, read=True, notify=True)
    aioble.register_services(service)

    multitest.globals(BDADDR=aioble.config("mac"))
    multitest.next()

    # Wait for central to connect to us.
    print("advertise")
    connection = await aioble.advertise(
        20_000, adv_data=b"\x02\x01\x06\x04\xffMPY", timeout_ms=TIMEOUT_MS
    )
    print("connected")

    # Send a subscribed-write (but client isn't subscribed, won't send anything).
    multitest.wait("discovery")
    await asyncio.sleep_ms(100)
    characteristic.write("before-subscribe", send_update=True)

    # Send a subscribed-write (now client is subscribed, client should get notified).
    multitest.wait("subscribed")
    await asyncio.sleep_ms(100)
    characteristic.write("after-subscribe", send_update=True)

    # Send a subscribed-write (now client is unsubscribed, won't send anything).
    multitest.wait("unsubscribed")
    await asyncio.sleep_ms(100)
    characteristic.write("after-unsubscribe", send_update=True)

    # Send 5 direct notifications.
    multitest.wait("start-direct")
    for i in range(5):
        # Send 1 notification each time, except for 3 quick notifications the third time.
        # The client should only see the last one.
        for j in range(3 if i == 2 else 1):
            if j > 0:
                await asyncio.sleep_ms(100)
            msg = "direct-{}-{}".format(i, j)
            print("notify", msg)
            characteristic.notify(connection, msg)

        # Tell client to wait for notification.
        multitest.broadcast("notified")
        # Wait until client is ready for next notification.
        multitest.wait("next")

    # Wait for the central to disconnect.
    await connection.disconnected(timeout_ms=TIMEOUT_MS)
    print("disconnected")


def instance0():
    try:
        asyncio.run(instance0_task())
    finally:
        aioble.stop()


# Acting in central role.
async def instance1_task():
    multitest.next()

    # Connect to peripheral and then disconnect.
    print("connect")
    device = aioble.Device(*BDADDR)
    connection = await device.connect(timeout_ms=TIMEOUT_MS)

    # Discover characteristics.
    service = await connection.service(SERVICE_UUID)
    print("service", service.uuid)
    characteristic = await service.characteristic(CHAR_UUID)
    print("characteristic", characteristic.uuid)

    # Expect to not receive a notification (not subscribed).
    multitest.broadcast("discovery")
    try:
        await characteristic.notified(timeout_ms=500)
        print("fail")
        return
    except asyncio.TimeoutError:
        print("no notification")

    # Subscribe and expect a notification.
    await characteristic.subscribe(notify=True)
    multitest.broadcast("subscribed")
    value = await characteristic.notified()
    print("notified", value)

    # Unsubscribe, and expect not to receive a notification.
    await characteristic.subscribe(notify=False)
    multitest.broadcast("unsubscribed")
    try:
        await characteristic.notified(timeout_ms=500)
        print("fail")
        return
    except asyncio.TimeoutError:
        print("no notification")

    # Receive 5 notifications.
    multitest.broadcast("start-direct")
    for i in range(5):
        multitest.wait("notified")
        await asyncio.sleep_ms(200)
        value = await characteristic.notified()
        print("notified", value)

        # Expect that after receiving a notification we don't get another one
        # until we broadcast to the server.
        try:
            value = await characteristic.notified(timeout_ms=100)
            print("unexpected notify", value)
        except asyncio.TimeoutError:
            pass

        multitest.broadcast("next")

    # Disconnect from peripheral.
    print("disconnect")
    await connection.disconnect(timeout_ms=TIMEOUT_MS)
    print("disconnected")


def instance1():
    try:
        asyncio.run(instance1_task())
    finally:
        aioble.stop()



try:
    import framebuf
except ImportError:
    print("SKIP")
    raise SystemExit


def printbuf():
    print("--8<--")
    for y in range(h):
        for x in range(w):
            print("%02x" % buf[(x + y * w)], end="")
        print()
    print("-->8--")


w = 30
h = 30
buf = bytearray(w * h)
fbuf = framebuf.FrameBuffer(buf, w, h, framebuf.GS8)

# Outline
fbuf.fill(0)
fbuf.ellipse(15, 15, 12, 6, 0xFF, False)
printbuf()

# Fill
fbuf.fill(0)
fbuf.ellipse(15, 15, 6, 12, 0xAA, True)
printbuf()

# Outline and fill some different quadrant combos.
for m in (0, 0b0001, 0b0010, 0b0100, 0b1000, 0b1010):
    fbuf.fill(0)
    fbuf.ellipse(15, 15, 6, 12, 0xAA, False, m)
    printbuf()
    fbuf.fill(0)
    fbuf.ellipse(15, 15, 6, 12, 0xAA, True, m)
    printbuf()

# Draw ellipses that will go out of bounds at each of the edges.
for x, y in (
    (
        4,
        4,
    ),
    (
        26,
        4,
    ),
    (
        26,
        26,
    ),
    (
        4,
        26,
    ),
):
    fbuf.fill(0)
    fbuf.ellipse(x, y, 6, 12, 0xAA, False)
    printbuf()
    fbuf.fill(0)
    fbuf.ellipse(x, y, 6, 12, 0xAA, True)
    printbuf()

# Draw an ellipse with both radius 0
fbuf.fill(0)
fbuf.ellipse(15, 15, 0, 0, 0xFF, True)
printbuf()

# Draw an ellipse with both radius 0 out of bounds
fbuf.fill(0)
fbuf.ellipse(45, 45, 0, 0, 0xFF, True)
printbuf()

# Draw an ellipse with radius 0 and all sectors masked out
fbuf.fill(0)
fbuf.ellipse(15, 15, 0, 0, 0xFF, True, 0)
printbuf()



# Test for sending notifications to subscribed clients.

from micropython import const
import time, machine, bluetooth

TIMEOUT_MS = 5000

_IRQ_CENTRAL_CONNECT = const(1)
_IRQ_CENTRAL_DISCONNECT = const(2)
_IRQ_GATTS_WRITE = const(3)
_IRQ_PERIPHERAL_CONNECT = const(7)
_IRQ_PERIPHERAL_DISCONNECT = const(8)
_IRQ_GATTC_CHARACTERISTIC_RESULT = const(11)
_IRQ_GATTC_CHARACTERISTIC_DONE = const(12)
_IRQ_GATTC_DESCRIPTOR_RESULT = const(13)
_IRQ_GATTC_DESCRIPTOR_DONE = const(14)
_IRQ_GATTC_READ_RESULT = const(15)
_IRQ_GATTC_READ_DONE = const(16)
_IRQ_GATTC_WRITE_DONE = const(17)
_IRQ_GATTC_NOTIFY = const(18)
_IRQ_GATTC_INDICATE = const(19)

_CCCD_UUID = bluetooth.UUID(const(0x2902))

SERVICE_UUID = bluetooth.UUID("A5A5A5A5-FFFF-9999-1111-5A5A5A5A5A5A")
CHAR_UUID = bluetooth.UUID("00000000-1111-2222-3333-444444444444")
CHAR = (
    CHAR_UUID,
    bluetooth.FLAG_READ | bluetooth.FLAG_WRITE | bluetooth.FLAG_NOTIFY | bluetooth.FLAG_INDICATE,
)
SERVICE = (
    SERVICE_UUID,
    (CHAR,),
)
SERVICES = (SERVICE,)

waiting_events = {}


def irq(event, data):
    if event == _IRQ_CENTRAL_CONNECT:
        print("_IRQ_CENTRAL_CONNECT")
        waiting_events[event] = data[0]
    elif event == _IRQ_CENTRAL_DISCONNECT:
        print("_IRQ_CENTRAL_DISCONNECT")
    elif event == _IRQ_GATTS_WRITE:
        print("_IRQ_GATTS_WRITE", ble.gatts_read(data[-1]))
    elif event == _IRQ_PERIPHERAL_CONNECT:
        print("_IRQ_PERIPHERAL_CONNECT")
        waiting_events[event] = data[0]
    elif event == _IRQ_PERIPHERAL_DISCONNECT:
        print("_IRQ_PERIPHERAL_DISCONNECT")
    elif event == _IRQ_GATTC_CHARACTERISTIC_RESULT:
        # conn_handle, end_handle, value_handle, properties, uuid = data
        if data[-1] == CHAR_UUID:
            print("_IRQ_GATTC_CHARACTERISTIC_RESULT", data[-1])
            waiting_events[event] = (data[2], data[1])
        else:
            return
    elif event == _IRQ_GATTC_CHARACTERISTIC_DONE:
        print("_IRQ_GATTC_CHARACTERISTIC_DONE")
    elif event == _IRQ_GATTC_DESCRIPTOR_RESULT:
        # conn_handle, dsc_handle, uuid = data
        if data[-1] == _CCCD_UUID:
            print("_IRQ_GATTC_DESCRIPTOR_RESULT", data[-1])
            waiting_events[event] = data[1]
        else:
            return
    elif event == _IRQ_GATTC_DESCRIPTOR_DONE:
        print("_IRQ_GATTC_DESCRIPTOR_DONE")
    elif event == _IRQ_GATTC_READ_RESULT:
        print("_IRQ_GATTC_READ_RESULT", bytes(data[-1]))
    elif event == _IRQ_GATTC_READ_DONE:
        print("_IRQ_GATTC_READ_DONE", data[-1])
    elif event == _IRQ_GATTC_WRITE_DONE:
        print("_IRQ_GATTC_WRITE_DONE", data[-1])
    elif event == _IRQ_GATTC_NOTIFY:
        print("_IRQ_GATTC_NOTIFY", bytes(data[-1]))
    elif event == _IRQ_GATTC_INDICATE:
        print("_IRQ_GATTC_NOTIFY", bytes(data[-1]))

    if event not in waiting_events:
        waiting_events[event] = None


def wait_for_event(event, timeout_ms):
    t0 = time.ticks_ms()
    while time.ticks_diff(time.ticks_ms(), t0) < timeout_ms:
        if event in waiting_events:
            return waiting_events.pop(event)
        machine.idle()
    raise ValueError("Timeout waiting for {}".format(event))


# Acting in peripheral role.
def instance0():
    multitest.globals(BDADDR=ble.config("mac"))
    ((char_handle,),) = ble.gatts_register_services(SERVICES)
    print("gap_advertise")
    ble.gap_advertise(20_000, b"\x02\x01\x06\x04\xffMPY")  # \x04\x09MPY
    multitest.next()
    try:
        # Write initial characteristic value (will be read by client).
        ble.gatts_write(char_handle, "periph0")  ###

        # Wait for central to connect to us.
        conn_handle = wait_for_event(_IRQ_CENTRAL_CONNECT, TIMEOUT_MS * 10)

        # A
        # Wait for a write to the characteristic from the central (to synchronise).
        wait_for_event(_IRQ_GATTS_WRITE, TIMEOUT_MS)
        print("sync A")
        # This should be local-only.
        ble.gatts_write(char_handle, "periph1")
        time.sleep_ms(100)
        # Update local-only, then force notify.
        ble.gatts_write(char_handle, "periph2")
        ble.gatts_notify(conn_handle, char_handle)  ###
        time.sleep_ms(100)
        # Update local and notify subscribers. No notification should be sent.
        ble.gatts_write(char_handle, "periph3", True)
        time.sleep_ms(100)
        multitest.broadcast("A")

        # B
        # Synchronise with the client (which should now be subscribed for notify).
        wait_for_event(_IRQ_GATTS_WRITE, TIMEOUT_MS)
        print("sync B")
        # This should be local-only (send_update=False).
        ble.gatts_write(char_handle, "periph4", False)
        time.sleep_ms(100)
        # This should notify the subscribed client.
        ble.gatts_write(char_handle, "periph5", True)  ###
        time.sleep_ms(100)
        multitest.broadcast("B")

        # C
        # Synchronise with the client (which should now be subscribed for indicate).
        wait_for_event(_IRQ_GATTS_WRITE, TIMEOUT_MS)
        print("sync C")
        # This should be local-only (send_update=False).
        ble.gatts_write(char_handle, "periph6", False)
        time.sleep_ms(100)
        # This should indicate the subscribed client.
        ble.gatts_write(char_handle, "periph7", True)  ###
        time.sleep_ms(100)
        multitest.broadcast("C")

        # D
        # Synchronise with the client (which should now be unsubscribed).
        wait_for_event(_IRQ_GATTS_WRITE, TIMEOUT_MS)
        print("sync D")
        # This should be local-only (send_update=False).
        ble.gatts_write(char_handle, "periph8", False)
        time.sleep_ms(100)
        # This should be local-only (no more subscribers).
        ble.gatts_write(char_handle, "periph9", True)
        time.sleep_ms(100)
        # Update local-only, then another force notify.
        ble.gatts_write(char_handle, "periph10")
        ble.gatts_notify(conn_handle, char_handle)  ###
        time.sleep_ms(100)
        multitest.broadcast("D")

        # Wait for the central to disconnect.
        wait_for_event(_IRQ_CENTRAL_DISCONNECT, TIMEOUT_MS)
    finally:
        ble.active(0)


# Acting in central role.
def instance1():
    multitest.next()
    try:
        # Connect to peripheral and then disconnect.
        print("gap_connect")
        ble.gap_connect(*BDADDR)
        conn_handle = wait_for_event(_IRQ_PERIPHERAL_CONNECT, TIMEOUT_MS)

        # Discover characteristics.
        ble.gattc_discover_characteristics(conn_handle, 1, 65535)
        value_handle, end_handle = wait_for_event(_IRQ_GATTC_CHARACTERISTIC_RESULT, TIMEOUT_MS)
        wait_for_event(_IRQ_GATTC_CHARACTERISTIC_DONE, TIMEOUT_MS)

        # Discover CCCD.
        ble.gattc_discover_descriptors(conn_handle, value_handle, end_handle)
        cccd_handle = wait_for_event(_IRQ_GATTC_DESCRIPTOR_RESULT, TIMEOUT_MS)
        wait_for_event(_IRQ_GATTC_DESCRIPTOR_DONE, TIMEOUT_MS)

        # Issue read of characteristic, should get initial value.
        print("gattc_read")
        ble.gattc_read(conn_handle, value_handle)
        wait_for_event(_IRQ_GATTC_READ_RESULT, TIMEOUT_MS)

        # While the four states are active, all incoming notifications
        # and indications will be printed by the event handler. We
        # should only expect to see certain ones.

        # Start unsubscribed.
        # Write to the characteristic (triggers A).
        print("gattc_write")
        ble.gattc_write(conn_handle, value_handle, "central0", 1)
        wait_for_event(_IRQ_GATTC_WRITE_DONE, TIMEOUT_MS)
        # Wait for A to complete.
        multitest.wait("A")

        # Subscribe for notify.
        ble.gattc_write(conn_handle, cccd_handle, b"\x01\x00", 1)
        wait_for_event(_IRQ_GATTC_WRITE_DONE, TIMEOUT_MS)
        # Write to the characteristic (triggers B).
        print("gattc_write")
        ble.gattc_write(conn_handle, value_handle, "central1", 1)
        wait_for_event(_IRQ_GATTC_WRITE_DONE, TIMEOUT_MS)
        # Wait for B to complete.
        multitest.wait("B")

        # Subscribe for indicate.
        ble.gattc_write(conn_handle, cccd_handle, b"\x02\x00", 1)
        wait_for_event(_IRQ_GATTC_WRITE_DONE, TIMEOUT_MS)
        # Write to the characteristic (triggers C).
        print("gattc_write")
        ble.gattc_write(conn_handle, value_handle, "central2", 1)
        wait_for_event(_IRQ_GATTC_WRITE_DONE, TIMEOUT_MS)
        # Wait for C to complete.
        multitest.wait("C")

        # Unsubscribe.
        ble.gattc_write(conn_handle, cccd_handle, b"\x00\x00", 1)
        wait_for_event(_IRQ_GATTC_WRITE_DONE, TIMEOUT_MS)
        # Write to the characteristic (triggers D).
        print("gattc_write")
        ble.gattc_write(conn_handle, value_handle, "central3", 1)
        wait_for_event(_IRQ_GATTC_WRITE_DONE, TIMEOUT_MS)
        # Wait for D to complete.
        multitest.wait("D")

        # Disconnect from peripheral.
        print("gap_disconnect:", ble.gap_disconnect(conn_handle))
        wait_for_event(_IRQ_PERIPHERAL_DISCONNECT, TIMEOUT_MS)
    finally:
        ble.active(0)


ble = bluetooth.BLE()
ble.active(1)
ble.irq(irq)



# test equality for classes/instances to other types

class A:
    pass

class B:
    pass

class C(A):
    pass

print(A == None)
print(None == A)

print(A == A)
print(A() == A)
print(A() == A())

print(A == B)
print(A() == B)
print(A() == B())

print(A == C)
print(A() == C)
print(A() == C())



"""NRF24L01 driver for MicroPython
"""

from micropython import const
import utime

# nRF24L01+ registers
CONFIG = const(0x00)
EN_RXADDR = const(0x02)
SETUP_AW = const(0x03)
SETUP_RETR = const(0x04)
RF_CH = const(0x05)
RF_SETUP = const(0x06)
STATUS = const(0x07)
RX_ADDR_P0 = const(0x0A)
TX_ADDR = const(0x10)
RX_PW_P0 = const(0x11)
FIFO_STATUS = const(0x17)
DYNPD = const(0x1C)

# CONFIG register
EN_CRC = const(0x08)  # enable CRC
CRCO = const(0x04)  # CRC encoding scheme; 0=1 byte, 1=2 bytes
PWR_UP = const(0x02)  # 1=power up, 0=power down
PRIM_RX = const(0x01)  # RX/TX control; 0=PTX, 1=PRX

# RF_SETUP register
POWER_0 = const(0x00)  # -18 dBm
POWER_1 = const(0x02)  # -12 dBm
POWER_2 = const(0x04)  # -6 dBm
POWER_3 = const(0x06)  # 0 dBm
SPEED_1M = const(0x00)
SPEED_2M = const(0x08)
SPEED_250K = const(0x20)

# STATUS register
RX_DR = const(0x40)  # RX data ready; write 1 to clear
TX_DS = const(0x20)  # TX data sent; write 1 to clear
MAX_RT = const(0x10)  # max retransmits reached; write 1 to clear

# FIFO_STATUS register
RX_EMPTY = const(0x01)  # 1 if RX FIFO is empty

# constants for instructions
R_RX_PL_WID = const(0x60)  # read RX payload width
R_RX_PAYLOAD = const(0x61)  # read RX payload
W_TX_PAYLOAD = const(0xA0)  # write TX payload
FLUSH_TX = const(0xE1)  # flush TX FIFO
FLUSH_RX = const(0xE2)  # flush RX FIFO
NOP = const(0xFF)  # use to read STATUS register


class NRF24L01:
    def __init__(self, spi, cs, ce, channel=46, payload_size=16):
        assert payload_size <= 32

        self.buf = bytearray(1)

        # store the pins
        self.spi = spi
        self.cs = cs
        self.ce = ce

        # init the SPI bus and pins
        self.init_spi(4000000)

        # reset everything
        ce.init(ce.OUT, value=0)
        cs.init(cs.OUT, value=1)

        self.payload_size = payload_size
        self.pipe0_read_addr = None
        utime.sleep_ms(5)

        # set address width to 5 bytes and check for device present
        self.reg_write(SETUP_AW, 0b11)
        if self.reg_read(SETUP_AW) != 0b11:
            raise OSError("nRF24L01+ Hardware not responding")

        # disable dynamic payloads
        self.reg_write(DYNPD, 0)

        # auto retransmit delay: 1750us
        # auto retransmit count: 8
        self.reg_write(SETUP_RETR, (6 << 4) | 8)

        # set rf power and speed
        self.set_power_speed(POWER_3, SPEED_250K)  # Best for point to point links

        # init CRC
        self.set_crc(2)

        # clear status flags
        self.reg_write(STATUS, RX_DR | TX_DS | MAX_RT)

        # set channel
        self.set_channel(channel)

        # flush buffers
        self.flush_rx()
        self.flush_tx()

    def init_spi(self, baudrate):
        try:
            master = self.spi.MASTER
        except AttributeError:
            self.spi.init(baudrate=baudrate, polarity=0, phase=0)
        else:
            self.spi.init(master, baudrate=baudrate, polarity=0, phase=0)

    def reg_read(self, reg):
        self.cs(0)
        self.spi.readinto(self.buf, reg)
        self.spi.readinto(self.buf)
        self.cs(1)
        return self.buf[0]

    def reg_write_bytes(self, reg, buf):
        self.cs(0)
        self.spi.readinto(self.buf, 0x20 | reg)
        self.spi.write(buf)
        self.cs(1)
        return self.buf[0]

    def reg_write(self, reg, value):
        self.cs(0)
        self.spi.readinto(self.buf, 0x20 | reg)
        ret = self.buf[0]
        self.spi.readinto(self.buf, value)
        self.cs(1)
        return ret

    def flush_rx(self):
        self.cs(0)
        self.spi.readinto(self.buf, FLUSH_RX)
        self.cs(1)

    def flush_tx(self):
        self.cs(0)
        self.spi.readinto(self.buf, FLUSH_TX)
        self.cs(1)

    # power is one of POWER_x defines; speed is one of SPEED_x defines
    def set_power_speed(self, power, speed):
        setup = self.reg_read(RF_SETUP) & 0b11010001
        self.reg_write(RF_SETUP, setup | power | speed)

    # length in bytes: 0, 1 or 2
    def set_crc(self, length):
        config = self.reg_read(CONFIG) & ~(CRCO | EN_CRC)
        if length == 0:
            pass
        elif length == 1:
            config |= EN_CRC
        else:
            config |= EN_CRC | CRCO
        self.reg_write(CONFIG, config)

    def set_channel(self, channel):
        self.reg_write(RF_CH, min(channel, 125))

    # address should be a bytes object 5 bytes long
    def open_tx_pipe(self, address):
        assert len(address) == 5
        self.reg_write_bytes(RX_ADDR_P0, address)
        self.reg_write_bytes(TX_ADDR, address)
        self.reg_write(RX_PW_P0, self.payload_size)

    # address should be a bytes object 5 bytes long
    # pipe 0 and 1 have 5 byte address
    # pipes 2-5 use same 4 most-significant bytes as pipe 1, plus 1 extra byte
    def open_rx_pipe(self, pipe_id, address):
        assert len(address) == 5
        assert 0 <= pipe_id <= 5
        if pipe_id == 0:
            self.pipe0_read_addr = address
        if pipe_id < 2:
            self.reg_write_bytes(RX_ADDR_P0 + pipe_id, address)
        else:
            self.reg_write(RX_ADDR_P0 + pipe_id, address[0])
        self.reg_write(RX_PW_P0 + pipe_id, self.payload_size)
        self.reg_write(EN_RXADDR, self.reg_read(EN_RXADDR) | (1 << pipe_id))

    def start_listening(self):
        self.reg_write(CONFIG, self.reg_read(CONFIG) | PWR_UP | PRIM_RX)
        self.reg_write(STATUS, RX_DR | TX_DS | MAX_RT)

        if self.pipe0_read_addr is not None:
            self.reg_write_bytes(RX_ADDR_P0, self.pipe0_read_addr)

        self.flush_rx()
        self.flush_tx()
        self.ce(1)
        utime.sleep_us(130)

    def stop_listening(self):
        self.ce(0)
        self.flush_tx()
        self.flush_rx()

    # returns True if any data available to recv
    def any(self):
        return not bool(self.reg_read(FIFO_STATUS) & RX_EMPTY)

    def recv(self):
        # get the data
        self.cs(0)
        self.spi.readinto(self.buf, R_RX_PAYLOAD)
        buf = self.spi.read(self.payload_size)
        self.cs(1)
        # clear RX ready flag
        self.reg_write(STATUS, RX_DR)

        return buf

    # blocking wait for tx complete
    def send(self, buf, timeout=500):
        self.send_start(buf)
        start = utime.ticks_ms()
        result = None
        while result is None and utime.ticks_diff(utime.ticks_ms(), start) < timeout:
            result = self.send_done()  # 1 == success, 2 == fail
        if result == 2:
            raise OSError("send failed")

    # non-blocking tx
    def send_start(self, buf):
        # power up
        self.reg_write(CONFIG, (self.reg_read(CONFIG) | PWR_UP) & ~PRIM_RX)
        utime.sleep_us(150)
        # send the data
        self.cs(0)
        self.spi.readinto(self.buf, W_TX_PAYLOAD)
        self.spi.write(buf)
        if len(buf) < self.payload_size:
            self.spi.write(b"\x00" * (self.payload_size - len(buf)))  # pad out data
        self.cs(1)

        # enable the chip so it can send the data
        self.ce(1)
        utime.sleep_us(15)  # needs to be >10us
        self.ce(0)

    # returns None if send still in progress, 1 for success, 2 for fail
    def send_done(self):
        if not (self.reg_read(STATUS) & (TX_DS | MAX_RT)):
            return None  # tx not finished

        # either finished or failed: get and clear status flags, power down
        status = self.reg_write(STATUS, RX_DR | TX_DS | MAX_RT)
        self.reg_write(CONFIG, self.reg_read(CONFIG) & ~PWR_UP)
        return 1 if status & TX_DS else 2



# Test descriptor discovery/read/write from both GATTS and GATTC.

import sys

# ruff: noqa: E402
sys.path.append("")

from micropython import const
import machine
import time

import asyncio
import aioble
import bluetooth

TIMEOUT_MS = 5000

SERVICE_UUID = bluetooth.UUID("A5A5A5A5-FFFF-9999-1111-5A5A5A5A5A5A")
CHAR1_UUID = bluetooth.UUID("00000000-1111-2222-3333-444444444444")
CHAR1_DESC1_UUID = bluetooth.UUID("00000000-1111-2222-3333-444444444445")
CHAR1_DESC2_UUID = bluetooth.UUID("00000000-1111-2222-3333-444444444446")
CHAR2_UUID = bluetooth.UUID("00000000-1111-2222-3333-555555555555")
CHAR3_UUID = bluetooth.UUID("00000000-1111-2222-3333-666666666666")
CHAR3_DESC1_UUID = bluetooth.UUID("00000000-1111-2222-3333-666666666667")


# Acting in peripheral role.
async def instance0_task():
    service = aioble.Service(SERVICE_UUID)
    char1 = aioble.Characteristic(
        service, CHAR1_UUID, read=True, write=True, notify=True, indicate=True
    )
    char1_desc1 = aioble.Descriptor(char1, CHAR1_DESC1_UUID, read=True)
    char1_desc1.write("char1_desc1")
    char1_desc2 = aioble.Descriptor(char1, CHAR1_DESC2_UUID, read=True, write=True)
    char1_desc2.write("char1_desc2")
    aioble.Characteristic(service, CHAR2_UUID, read=True, write=True, notify=True, indicate=True)
    char3 = aioble.Characteristic(
        service, CHAR3_UUID, read=True, write=True, notify=True, indicate=True
    )
    char3_desc1 = aioble.Descriptor(char3, CHAR3_DESC1_UUID, read=True)
    char3_desc1.write("char3_desc1")
    aioble.register_services(service)

    multitest.globals(BDADDR=aioble.config("mac"))
    multitest.next()

    # Wait for central to connect to us.
    print("advertise")
    connection = await aioble.advertise(
        20_000, adv_data=b"\x02\x01\x06\x04\xffMPY", timeout_ms=TIMEOUT_MS
    )
    print("connected")

    await char1_desc2.written()
    print("written", char1_desc2.read())

    # Wait for the central to disconnect.
    await connection.disconnected(timeout_ms=TIMEOUT_MS)
    print("disconnected")


def instance0():
    try:
        asyncio.run(instance0_task())
    finally:
        aioble.stop()


# Acting in central role.
async def instance1_task():
    multitest.next()

    # Connect to peripheral and then disconnect.
    print("connect")
    device = aioble.Device(*BDADDR)
    connection = await device.connect(timeout_ms=TIMEOUT_MS)

    service = await connection.service(SERVICE_UUID)
    print("service", service.uuid)

    # Discover characteristics.
    uuids = []
    async for char in service.characteristics():
        uuids.append(char.uuid)
    print("found", sorted(uuids))

    char1 = await service.characteristic(CHAR1_UUID)
    print("char1", char1.uuid)

    uuids = []
    async for desc in char1.descriptors():
        uuids.append(desc.uuid)
    print("char1 descs", sorted(uuids))

    char1_desc1 = await char1.descriptor(CHAR1_DESC1_UUID)
    print("char1 desc1", char1_desc1.uuid)
    print("read", await char1_desc1.read())

    char1_desc2 = await char1.descriptor(CHAR1_DESC2_UUID)
    print("char1 desc2", char1_desc2.uuid)
    print("read", await char1_desc2.read())
    print("write")
    await char1_desc2.write("update")

    char2 = await service.characteristic(CHAR2_UUID)
    print("char2", char2.uuid)

    uuids = []
    async for desc in char2.descriptors():
        uuids.append(desc.uuid)
    print("char2 descs", sorted(uuids))

    char3 = await service.characteristic(CHAR3_UUID)
    print("char3", char3.uuid)

    uuids = []
    async for desc in char3.descriptors():
        uuids.append(desc.uuid)
    print("char3 descs", sorted(uuids))

    char3_desc1 = await char3.descriptor(CHAR3_DESC1_UUID)
    print("char3 desc1", char3_desc1.uuid)
    print("read", await char3_desc1.read())

    # Disconnect from peripheral.
    print("disconnect")
    await connection.disconnect(timeout_ms=TIMEOUT_MS)
    print("disconnected")


def instance1():
    try:
        asyncio.run(instance1_task())
    finally:
        aioble.stop()



# Test get_event_loop()

try:
    import asyncio
except ImportError:
    print("SKIP")
    raise SystemExit

# CPython 3.12 deprecated calling get_event_loop() when there is no current event
# loop, so to make this test run on CPython requires setting the event loop.
if hasattr(asyncio, "set_event_loop"):
    asyncio.set_event_loop(asyncio.new_event_loop())


async def main():
    print("start")
    await asyncio.sleep(0.01)
    print("end")


loop = asyncio.get_event_loop()
loop.run_until_complete(main())



# test ffi float support
try:
    import ffi
except ImportError:
    print("SKIP")
    raise SystemExit


def ffi_open(names):
    err = None
    for n in names:
        try:
            mod = ffi.open(n)
            return mod
        except OSError as e:
            err = e
    raise err


libm = ffi_open(("libm.so", "libm.so.6", "libc.so.0", "libc.so.6", "libc.dylib"))

# Some libc's implement tgammaf as header macro with tgamma(), so don't assume
# it'll be in library.
try:
    tgammaf = libm.func("f", "tgammaf", "f")
except OSError:
    print("SKIP")
    raise SystemExit

for fun in (tgammaf,):
    for val in (0.5, 1, 1.0, 1.5, 4, 4.0):
        print("%.6f" % fun(val))



class MyGen:

    def __init__(self):
        self.v = 0

    def __iter__(self):
        return self

    def __next__(self):
        self.v += 1
        if self.v > 5:
            raise StopIteration
        return self.v

def gen():
    yield from MyGen()

def gen2():
    yield from gen()

print(list(gen()))
print(list(gen2()))


class Incrementer:

    def __iter__(self):
        return self

    def __next__(self):
        return self.send(None)

    def send(self, val):
        if val is None:
            return "Incrementer initialized"
        return val + 1

def gen3():
    yield from Incrementer()

g = gen3()
print(next(g))
print(g.send(5))
print(g.send(100))


#
# Test proper handling of StopIteration vs other exceptions
#
class MyIter:
    def __iter__(self):
        return self
    def __next__(self):
        raise StopIteration(42)

def gen4():
    global ret
    ret = yield from MyIter()
    1//0

ret = None
try:
    print(list(gen4()))
except ZeroDivisionError:
    print("ZeroDivisionError")
print(ret)



#!/usr/bin/env python3
#
# This file is part of the MicroPython project, http://micropython.org/
#
# The MIT License (MIT)
#
# Copyright (c) 2020 Damien P. George
# Copyright (c) 2020 Jim Mussared
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import argparse
import glob
import itertools
import os
import re
import subprocess

# Relative to top-level repo dir.
PATHS = [
    "drivers/ninaw10/*.[ch]",
    "extmod/*.[ch]",
    "extmod/btstack/*.[ch]",
    "extmod/nimble/*.[ch]",
    "lib/mbedtls_errors/tester.c",
    "shared/netutils/*.[ch]",
    "shared/timeutils/*.[ch]",
    "shared/runtime/*.[ch]",
    "shared/tinyusb/*.[ch]",
    "mpy-cross/*.[ch]",
    "ports/**/*.[ch]",
    "py/*.[ch]",
]

EXCLUSIONS = [
    # The cc3200 port is not fully formatted yet.
    "ports/cc3200/*/*.[ch]",
    # ESP-IDF downloads 3rd party code.
    "ports/esp32/managed_components/*",
    # The nrf port is not fully formatted yet.
    "ports/nrf/boards/*.[ch]",
    "ports/nrf/device/*.[ch]",
    "ports/nrf/drivers/*.[ch]",
    "ports/nrf/modules/ble/*.[ch]",
    "ports/nrf/modules/board/*.[ch]",
    "ports/nrf/modules/music/*.[ch]",
    "ports/nrf/modules/ubluepy/*.[ch]",
    "ports/nrf/modules/os/*.[ch]",
    # STM32 USB dev/host code is mostly 3rd party.
    "ports/stm32/usbdev/**/*.[ch]",
    "ports/stm32/usbhost/**/*.[ch]",
    # STM32 build includes generated Python code.
    "ports/*/build*",
]

# Path to repo top-level dir.
TOP = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))

UNCRUSTIFY_CFG = os.path.join(TOP, "tools/uncrustify.cfg")


def list_files(paths, exclusions=None, prefix=""):
    files = set()
    for pattern in paths:
        files.update(glob.glob(os.path.join(prefix, pattern), recursive=True))
    for pattern in exclusions or []:
        files.difference_update(glob.fnmatch.filter(files, os.path.join(prefix, pattern)))
    return sorted(files)


def fixup_c(filename):
    # Read file.
    with open(filename) as f:
        lines = f.readlines()

    # Write out file with fixups.
    with open(filename, "w", newline="") as f:
        dedent_stack = []
        while lines:
            # Get next line.
            l = lines.pop(0)

            # Dedent #'s to match indent of following line (not previous line).
            m = re.match(r"( +)#(if |ifdef |ifndef |elif |else|endif)", l)
            if m:
                indent = len(m.group(1))
                directive = m.group(2)
                if directive in ("if ", "ifdef ", "ifndef "):
                    l_next = lines[0]
                    indent_next = len(re.match(r"( *)", l_next).group(1))
                    if indent - 4 == indent_next and re.match(r" +(} else |case )", l_next):
                        # This #-line (and all associated ones) needs dedenting by 4 spaces.
                        l = l[4:]
                        dedent_stack.append(indent - 4)
                    else:
                        # This #-line does not need dedenting.
                        dedent_stack.append(-1)
                else:
                    if dedent_stack[-1] >= 0:
                        # This associated #-line needs dedenting to match the #if.
                        indent_diff = indent - dedent_stack[-1]
                        assert indent_diff >= 0
                        l = l[indent_diff:]
                    if directive == "endif":
                        dedent_stack.pop()

            # Write out line.
            f.write(l)

        assert not dedent_stack, filename


def main():
    cmd_parser = argparse.ArgumentParser(description="Auto-format C and Python files.")
    cmd_parser.add_argument("-c", action="store_true", help="Format C code only")
    cmd_parser.add_argument("-p", action="store_true", help="Format Python code only")
    cmd_parser.add_argument("-v", action="store_true", help="Enable verbose output")
    cmd_parser.add_argument(
        "-f",
        action="store_true",
        help="Filter files provided on the command line against the default list of files to check.",
    )
    cmd_parser.add_argument("files", nargs="*", help="Run on specific globs")
    args = cmd_parser.parse_args()

    # Setting only one of -c or -p disables the other. If both or neither are set, then do both.
    format_c = args.c or not args.p
    format_py = args.p or not args.c

    # Expand the globs passed on the command line, or use the default globs above.
    files = []
    if args.files:
        files = list_files(args.files)
        if args.f:
            # Filter against the default list of files. This is a little fiddly
            # because we need to apply both the inclusion globs given in PATHS
            # as well as the EXCLUSIONS, and use absolute paths
            files = set(os.path.abspath(f) for f in files)
            all_files = set(list_files(PATHS, EXCLUSIONS, TOP))
            if args.v:  # In verbose mode, log any files we're skipping
                for f in files - all_files:
                    print("Not checking: {}".format(f))
            files = list(files & all_files)
    else:
        files = list_files(PATHS, EXCLUSIONS, TOP)

    # Run tool on N files at a time (to avoid making the command line too long).
    def batch(cmd, N=200):
        files_iter = iter(files)
        while True:
            file_args = list(itertools.islice(files_iter, N))
            if not file_args:
                break
            subprocess.check_call(cmd + file_args)

    # Format C files with uncrustify.
    if format_c:
        command = ["uncrustify", "-c", UNCRUSTIFY_CFG, "-lC", "--no-backup"]
        if not args.v:
            command.append("-q")
        batch(command)
        for file in files:
            fixup_c(file)

    # Format Python files with "ruff format" (using config in pyproject.toml).
    if format_py:
        command = ["ruff", "format"]
        if args.v:
            command.append("-v")
        else:
            command.append("-q")
        command.append(".")
        subprocess.check_call(command, cwd=TOP)


if __name__ == "__main__":
    main()



# test calling a subclass of a native class that supports calling

# For this test we need a native class that can be subclassed (has make_new)
# and is callable (has call).  The only one available is machine.Signal, which
# in turns needs PinBase.
try:
    import machine
    machine.PinBase
    machine.Signal
except:
    print("SKIP")
    raise SystemExit

class Pin(machine.PinBase):
    #def __init__(self):
    #    self.v = 0

    def value(self, v=None):
        return 42

class MySignal(machine.Signal):
    pass

s = MySignal(Pin())

# apply call to the subclass, which should call the native base
print(s())



# Test the limits of bytecode generation.

body = " with f()()() as a:\n  try:\n   f()()()\n  except Exception:\n   pass\n"

# Test overflow of jump offset.
# Print results at the end in case an intermediate value of n fails with MemoryError.
results = []
for n in (433, 432, 431, 430):
    try:
        exec("cond = 0\nif cond:\n" + body * n + "else:\n print('cond false')\n")
        results.append((n, "ok"))
    except MemoryError:
        print("SKIP")
        raise SystemExit
    except RuntimeError:
        results.append((n, "RuntimeError"))
print(results)

# Test changing size of code info (source line/bytecode mapping) due to changing
# bytecode size in the final passes.  This test is very specific to how the
# code info is encoded, and how jump offsets shrink in the final passes.  This
# test should fail if the bytecode emitter doesn't correctly handle shrinking of
# the code info section.
exec(
    """
x = 0
if x:
"""
    + body * 13
    + """
x = [1 if x else 123]






print(x)
"""
)



import os
from machine import UART

machine = os.uname().machine
if "EK-RA6M2" in machine:
    # 0, 7, 9
    uart_ids = (0, 7, 9)
    try_id = 7
    try_s = "UART(7, baudrate=115200, bits=8, parity=None, stop=1, tx=P401, rx=P402, flow=0, rxbuf=259, timeout=0, timeout_char=2)"
elif "RA4M1 CLICKER" in machine:
    # 0, 1
    uart_ids = (0, 1)
    try_id = 0
    try_s = "UART(0, baudrate=115200, bits=8, parity=None, stop=1, tx=P411, rx=P410, flow=0, rxbuf=259, timeout=0, timeout_char=2)"
elif "EK-RA4M1" in machine:
    # 0, 1, 2, 9
    # vector for 9 is not registered
    uart_ids = (0, 1, 2)
    try_id = 1
    try_s = "UART(1, baudrate=115200, bits=8, parity=None, stop=1, tx=P401, rx=P402, flow=0, rxbuf=259, timeout=0, timeout_char=2)"
elif "EK-RA4W1" in machine:
    # 0, 1, 4, 9
    # 0 is disabled.
    uart_ids = (1, 4, 9)
    try_id = 9
    try_s = "UART(9, baudrate=115200, bits=8, parity=None, stop=1, tx=P109, rx=P110, flow=0, rxbuf=259, timeout=0, timeout_char=2)"
elif "EK-RA6M1" in machine:
    # 0, 1, 2, 3, 4, 8, 9
    # 1/3/4/9 are disabled
    uart_ids = (0, 2, 8)
    try_id = 8
    try_s = "UART(8, baudrate=115200, bits=8, parity=None, stop=1, tx=P105, rx=P104, flow=0, rxbuf=259, timeout=0, timeout_char=2)"
else:
    print("SKIP")
    raise SystemExit

for ch in uart_ids:
    try:
        uart = UART(ch)
    except ValueError:
        print("ValueError", bus)

rxbuf_size = 259

ch = UART(try_id, rxbuf=rxbuf_size)
if str(ch) == try_s:
    print("OK")
else:
    print("NG")
    print("exp: " + try_s)
    print("out: " + str(ch))



# Test basic, stand-alone TCP socket functionality

try:
    import socket, errno
except ImportError:
    print("SKIP")
    raise SystemExit

# recv() on a fresh socket should raise ENOTCONN
s = socket.socket()
try:
    s.recv(1)
except OSError as er:
    print("ENOTCONN:", er.errno == errno.ENOTCONN)



import unittest


class Test(unittest.TestCase):
    def test_subtest_skip(self):
        for i in range(4):
            with self.subTest(i=i):
                print("sub test", i)
                if i == 2:
                    self.skipTest("skip 2")


if __name__ == "__main__":
    unittest.main()



# test __dict__ attribute of a built-in module
# see import/module_dict.py for the equivalent test on user modules

import sys

if not hasattr(sys, "__dict__"):
    print("SKIP")
    raise SystemExit


# dict of a built-in module (read-only)
print(type(sys.__dict__))
print(sys.__dict__["__name__"])



# test basic properties of exceptions

print(repr(IndexError()))
print(str(IndexError()))

print(str(IndexError("foo")))

a = IndexError(1, "test", [100, 200])
print(repr(a))
print(str(a))
print(a.args)

s = StopIteration()
print(s.value)
s = StopIteration(1, 2, 3)
print(s.value)

print(OSError().errno)
print(OSError(1, "msg").errno)



try:
    import hashlib
except ImportError:
    # This is neither uPy, nor cPy, so must be uPy with
    # hashlib module disabled.
    print("SKIP")
    raise SystemExit

try:
    hashlib.md5
except AttributeError:
    # MD5 is only available on some ports
    print("SKIP")
    raise SystemExit

md5 = hashlib.md5(b"hello")
md5.update(b"world")
print(md5.digest())



import pyb

if not hasattr(pyb, "DAC"):
    print("SKIP")
    raise SystemExit

dac = pyb.DAC(1)
print(dac)
dac.noise(100)
dac.triangle(100)
dac.write(0)
dac.write_timed(bytearray(10), 100, mode=pyb.DAC.NORMAL)
pyb.delay(20)
dac.write(0)

# test buffering arg
dac = pyb.DAC(1, buffering=True)
dac.write(0)



#!/usr/bin/env python3
#
# This file is part of the MicroPython project, http://micropython.org/
#
# The MIT License (MIT)
#
# Copyright (c) 2023 Jim Mussared
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Simple script to take the output printed when HCI_TRACE is enabled in
# unix/mpbthciport.c, extmod/btstack/btstack_hci_uart.c,
# extmod/nimble/hal/hal_uart.c and turn it into a .pcap file suitable for use
# with Wireshark.

import re
import sys
import struct

# pcap file header:
# typedef struct pcap_hdr_s {
#     guint32 magic_number;   /* magic number */
#     guint16 version_major;  /* major version number */
#     guint16 version_minor;  /* minor version number */
#     gint32  thiszone;       /* GMT to local correction */
#     guint32 sigfigs;        /* accuracy of timestamps */
#     guint32 snaplen;        /* max length of captured packets, in octets */
#     guint32 network;        /* data link type */
# } pcap_hdr_t;

# pcap record header
# typedef struct pcaprec_hdr_s {
#     guint32 ts_sec;         /* timestamp seconds */
#     guint32 ts_usec;        /* timestamp microseconds */
#     guint32 incl_len;       /* number of octets of packet saved in file */
#     guint32 orig_len;       /* actual length of packet */
# } pcaprec_hdr_t;

_LINKTYPE_BLUETOOTH_HCI_H4_WITH_PHDR = 201  # "!I" direction, followed by data
sys.stdout.buffer.write(
    struct.pack("!IHHiIII", 0xA1B2C3D4, 2, 4, 0, 0, 65535, _LINKTYPE_BLUETOOTH_HCI_H4_WITH_PHDR)
)

_DIR_CONTROLLER_TO_HOST = 1
_DIR_HOST_TO_CONTROLLER = 0

reassemble_timestamp = 0
reassemble_packet = bytearray()

with open(sys.argv[1], "r") as f:
    for line in f:
        line = line.strip()
        m = re.match("([<>]) \\[ *([0-9]+)\\] ([A-Fa-f0-9:]+)", line)
        if not m:
            continue

        timestamp = int(m.group(2))
        data = bytes.fromhex(m.group(3).replace(":", ""))

        if m.group(1) == "<":
            # Host to controller.
            # These are always complete.
            sys.stdout.buffer.write(
                struct.pack(
                    "!IIIII",
                    timestamp // 1000,
                    timestamp % 1000 * 1000,
                    len(data) + 4,
                    len(data) + 4,
                    _DIR_HOST_TO_CONTROLLER,
                )
            )
            sys.stdout.buffer.write(data)
        if m.group(1) == ">":
            # Controller to host.
            # Several of the sources print byte-at-a-time so need to reconstruct packets.

            if not reassemble_packet:
                # Use the timestamp of the first fragment.
                reassemble_timestamp = timestamp

            reassemble_packet.extend(data)

            if len(reassemble_packet) > 4:
                plen = 0
                if reassemble_packet[0] == 1:
                    # command
                    plen = 3 + reassemble_packet[3]
                elif reassemble_packet[0] == 2:
                    # acl
                    plen = 5 + reassemble_packet[3] + (reassemble_packet[4] << 8)
                elif reassemble_packet[0] == 4:
                    # event
                    plen = 3 + reassemble_packet[2]

                if len(reassemble_packet) >= plen:
                    # Got a complete packet.
                    data = reassemble_packet[0:plen]
                    reassemble_packet = reassemble_packet[plen:]
                    reassemble_timestamp = timestamp
                    sys.stdout.buffer.write(
                        struct.pack(
                            "!IIIII",
                            reassemble_timestamp // 1000,
                            reassemble_timestamp % 1000 * 1000,
                            len(data) + 4,
                            len(data) + 4,
                            _DIR_CONTROLLER_TO_HOST,
                        )
                    )
                    sys.stdout.buffer.write(data)

if reassemble_packet:
    print(
        "Error: Unknown byte in HCI stream. Remainder:",
        reassemble_packet.hex(":"),
        file=sys.stderr,
    )



try:
    from pyb import CAN

    CAN(2)
except (ImportError, ValueError):
    print("SKIP")
    raise SystemExit

# Testing rtr messages
bus2 = CAN(2, CAN.LOOPBACK)
while bus2.any(0):
    bus2.recv(0)
bus2.setfilter(0, CAN.LIST32, 0, (1, 2), rtr=(True, True), extframe=True)
bus2.setfilter(1, CAN.LIST32, 0, (3, 4), rtr=(True, False), extframe=True)
bus2.setfilter(2, CAN.MASK32, 0, (16, 16), rtr=(False,), extframe=True)
bus2.setfilter(2, CAN.MASK32, 0, (32, 32), rtr=(True,), extframe=True)

bus2.send("", 1, rtr=True, extframe=True)
print(bus2.recv(0))
bus2.send("", 2, rtr=True, extframe=True)
print(bus2.recv(0))
bus2.send("", 3, rtr=True, extframe=True)
print(bus2.recv(0))
bus2.send("", 4, rtr=True, extframe=True)
print(bus2.any(0))



try:
    import errno, os, vfs

    vfs.VfsFat
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit


class RAMFS:
    SEC_SIZE = 512

    def __init__(self, blocks):
        self.data = bytearray(blocks * self.SEC_SIZE)

    def readblocks(self, n, buf):
        # print("readblocks(%s, %x(%d))" % (n, id(buf), len(buf)))
        for i in range(len(buf)):
            buf[i] = self.data[n * self.SEC_SIZE + i]

    def writeblocks(self, n, buf):
        # print("writeblocks(%s, %x)" % (n, id(buf)))
        for i in range(len(buf)):
            self.data[n * self.SEC_SIZE + i] = buf[i]

    def ioctl(self, op, arg):
        # print("ioctl(%d, %r)" % (op, arg))
        if op == 4:  # MP_BLOCKDEV_IOCTL_BLOCK_COUNT
            return len(self.data) // self.SEC_SIZE
        if op == 5:  # MP_BLOCKDEV_IOCTL_BLOCK_SIZE
            return self.SEC_SIZE


try:
    bdev = RAMFS(50)
    vfs.VfsFat.mkfs(bdev)
except MemoryError:
    print("SKIP")
    raise SystemExit

fs = vfs.VfsFat(bdev)
vfs.mount(fs, "/ramdisk")
os.chdir("/ramdisk")

try:
    fs.mkdir("foo_dir")
except OSError as e:
    print(e.errno == errno.EEXIST)

try:
    fs.remove("foo_dir")
except OSError as e:
    print(e.errno == errno.EISDIR)

try:
    fs.remove("no_file.txt")
except OSError as e:
    print(e.errno == errno.ENOENT)

try:
    fs.rename("foo_dir", "/null/file")
except OSError as e:
    print(e.errno == errno.ENOENT)

# file in dir
with open("foo_dir/file-in-dir.txt", "w+t") as f:
    f.write("data in file")

with open("foo_dir/file-in-dir.txt", "r+b") as f:
    print(f.read())

with open("foo_dir/sub_file.txt", "w") as f:
    f.write("subdir file")

# directory not empty
try:
    fs.rmdir("foo_dir")
except OSError as e:
    print(e.errno == errno.EACCES)

# trim full path
fs.rename("foo_dir/file-in-dir.txt", "foo_dir/file.txt")
print(list(fs.ilistdir("foo_dir")))

fs.rename("foo_dir/file.txt", "moved-to-root.txt")
print(list(fs.ilistdir()))

# check that renaming to existing file will overwrite it
with open("temp", "w") as f:
    f.write("new text")
fs.rename("temp", "moved-to-root.txt")
print(list(fs.ilistdir()))
with open("moved-to-root.txt") as f:
    print(f.read())

# valid removes
fs.remove("foo_dir/sub_file.txt")
fs.rmdir("foo_dir")
print(list(fs.ilistdir()))

# disk full
try:
    bsize = fs.statvfs("/ramdisk")[0]
    free = fs.statvfs("/ramdisk")[2] + 1
    f = open("large_file.txt", "wb")
    f.write(bytearray(bsize * free))
except OSError as e:
    print("ENOSPC:", e.errno == 28)  # errno.ENOSPC
f.close()



# Test asyncio TCP stream closing then writing

try:
    import asyncio
except ImportError:
    print("SKIP")
    raise SystemExit

PORT = 8000


async def handle_connection(reader, writer):
    # Write data to ensure connection
    writer.write(b"x")
    await writer.drain()

    # Read, should return nothing
    print("read:", await reader.read(100))

    # Close connection
    print("close")
    writer.close()
    await writer.wait_closed()

    print("done")
    ev.set()


async def tcp_server():
    global ev
    ev = asyncio.Event()
    server = await asyncio.start_server(handle_connection, "0.0.0.0", PORT)
    print("server running")
    multitest.next()
    async with server:
        await asyncio.wait_for(ev.wait(), 10)


async def tcp_client():
    reader, writer = await asyncio.open_connection(IP, PORT)

    # Read data to ensure connection
    print("read:", await reader.read(1))

    # Close connection
    print("close")
    writer.close()
    await writer.wait_closed()

    # Try writing data to the closed connection
    print("write")
    try:
        writer.write(b"x")
        await writer.drain()
    except OSError:
        print("OSError")


def instance0():
    multitest.globals(IP=multitest.get_network_ip())
    asyncio.run(tcp_server())


def instance1():
    multitest.next()
    asyncio.run(tcp_client())



def gen1():
    yield 1
    yield 2

# Test that it's possible to close just created gen
g = gen1()
print(g.close())
try:
    next(g)
except StopIteration:
    print("StopIteration")

# Test that it's possible to close gen in progress
g = gen1()
print(next(g))
print(g.close())
try:
    next(g)
    print("No StopIteration")
except StopIteration:
    print("StopIteration")

# Test that it's possible to close finished gen
g = gen1()
print(list(g))
print(g.close())
try:
    next(g)
    print("No StopIteration")
except StopIteration:
    print("StopIteration")


# Throwing GeneratorExit in response to close() is ok
def gen2():
    try:
        yield 1
        yield 2
    except:
        print('raising GeneratorExit')
        raise GeneratorExit

g = gen2()
next(g)
print(g.close())

# Any other exception is propagated to the .close() caller
def gen3():
    try:
        yield 1
        yield 2
    except:
        raise ValueError

g = gen3()
next(g)
try:
    print(g.close())
except ValueError:
    print("ValueError")



"""
WLAN test for the CC3200 based boards.
"""

from network import WLAN
import os
import time
import testconfig

mch = os.uname().machine
if not "LaunchPad" in mch and not "WiPy" in mch:
    raise Exception("Board not supported!")


def wait_for_connection(wifi, timeout=10):
    while not wifi.isconnected() and timeout > 0:
        time.sleep(1)
        timeout -= 1
    if wifi.isconnected():
        print("Connected")
    else:
        print("Connection failed!")


wifi = WLAN(0, WLAN.STA)
print(wifi.mode() == WLAN.STA)
print(wifi.antenna() == WLAN.INT_ANT)

wifi = WLAN(mode=WLAN.AP)
print(wifi.mode() == WLAN.AP)
print(wifi.channel() == 1)
print(wifi.auth() == None)
print(wifi.antenna() == WLAN.INT_ANT)
wifi = WLAN(0, mode=WLAN.AP, ssid="test-wlan", auth=(WLAN.WPA, "123456abc"), channel=7)
print(wifi.mode() == WLAN.AP)
print(wifi.channel() == 7)
print(wifi.ssid() == "test-wlan")
print(wifi.auth() == (WLAN.WPA, "123456abc"))
print(wifi.antenna() == WLAN.INT_ANT)

wifi = WLAN(mode=WLAN.STA)
print(wifi.mode() == WLAN.STA)
time.sleep(5)  # this ensures a full network scan
scan_r = wifi.scan()
print(len(scan_r) > 3)
for net in scan_r:
    if net.ssid == testconfig.wlan_ssid:
        # test that the scan results contains the desired params
        print(len(net.bssid) == 6)
        print(net.channel == None)
        print(net.sec == testconfig.wlan_auth[0])
        print(net.rssi < 0)
        print("Network found")
        break

wifi.mode(WLAN.STA)
print(wifi.mode() == WLAN.STA)
wifi.channel(7)
print(wifi.channel() == 7)
wifi.ssid("t-wlan")
print(wifi.ssid() == "t-wlan")
wifi.auth(None)
print(wifi.auth() == None)
wifi.auth((WLAN.WEP, "11223344556677889900"))
print(wifi.auth() == (WLAN.WEP, "11223344556677889900"))
wifi.antenna(WLAN.INT_ANT)
print(wifi.antenna() == WLAN.INT_ANT)

wifi.antenna(WLAN.EXT_ANT)
print(wifi.antenna() == WLAN.EXT_ANT)
time.sleep(2)  # this ensures a full network scan
scan_r = wifi.scan()
print(len(scan_r) > 3)
for net in scan_r:
    if net.ssid == testconfig.wlan_ssid:
        print("Network found")
        break

wifi.antenna(WLAN.INT_ANT)
wifi.mode(WLAN.STA)
print(wifi.mode() == WLAN.STA)
wifi.connect(testconfig.wlan_ssid, auth=testconfig.wlan_auth, timeout=10000)
wait_for_connection(wifi)

wifi.ifconfig(config="dhcp")
wait_for_connection(wifi)
print("0.0.0.0" not in wifi.ifconfig())
wifi.ifconfig(0, ("192.168.178.109", "255.255.255.0", "192.168.178.1", "8.8.8.8"))
wait_for_connection(wifi)
print(wifi.ifconfig(0) == ("192.168.178.109", "255.255.255.0", "192.168.178.1", "8.8.8.8"))
wait_for_connection(wifi)

print(wifi.isconnected() == True)
wifi.disconnect()
print(wifi.isconnected() == False)

t0 = time.ticks_ms()
wifi.connect(testconfig.wlan_ssid, auth=testconfig.wlan_auth, timeout=0)
print(time.ticks_ms() - t0 < 500)

wifi.disconnect()
print(wifi.isconnected() == False)

# test init again
wifi.init(WLAN.AP, ssid="www.wipy.io", auth=None, channel=5, antenna=WLAN.INT_ANT)
print(wifi.mode() == WLAN.AP)

# get the current instance without re-init
wifi = WLAN()
print(wifi.mode() == WLAN.AP)
wifi = WLAN(0)
print(wifi.mode() == WLAN.AP)

# test the MAC address length
print(len(wifi.mac()) == 6)

# next ones MUST raise
try:
    wifi.init(mode=12345)
except:
    print("Exception")

try:
    wifi.init(1, mode=WLAN.AP)
except:
    print("Exception")

try:
    wifi.init(mode=WLAN.AP, ssid=None)
except:
    print("Exception")

try:
    wifi = WLAN(mode=WLAN.AP, channel=12)
except:
    print("Exception")

try:
    wifi.antenna(2)
except:
    print("Exception")

try:
    wifi.mode(10)
except:
    print("Exception")

try:
    wifi.ssid(
        "11111sdfasdfasdfasdf564sdf654asdfasdf123451245ssdgfsdf1111111111111111111111111234123412341234asdfasdf"
    )
except:
    print("Exception")

try:
    wifi.auth((0))
except:
    print("Exception")

try:
    wifi.auth((0, None))
except:
    print("Exception")

try:
    wifi.auth((10, 10))
except:
    print("Exception")

try:
    wifi.channel(0)
except:
    print("Exception")

try:
    wifi.ifconfig(1, "dhcp")
except:
    print("Exception")

try:
    wifi.ifconfig(config=())
except:
    print("Exception")



# See original soft_pwm.py for detailed comments.
import asyncio
from hwconfig import LED


async def pwm_cycle(led, duty, cycles):
    duty_off = 20 - duty
    for i in range(cycles):
        if duty:
            led.value(1)
            await asyncio.sleep_ms(duty)
        if duty_off:
            led.value(0)
            await asyncio.sleep_ms(duty_off)


async def fade_in_out(LED):
    while True:
        # Fade in
        for i in range(1, 21):
            await pwm_cycle(LED, i, 2)
        # Fade out
        for i in range(20, 0, -1):
            await pwm_cycle(LED, i, 2)


loop = asyncio.get_event_loop()
loop.run_until_complete(fade_in_out(LED))



#!/usr/bin/env python


from collections import defaultdict, namedtuple
import os
import re
import sys

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "../../../tools"))
import boardgen


# Must have matching entries in AF_FN_* enum in ../pin_defs_stm32.h
SUPPORTED_AF = {
    "TIM": ["CH1", "CH2", "CH3", "CH4", "CH1N", "CH2N", "CH3N", "CH1_ETR", "ETR", "BKIN"],
    "I2C": ["SDA", "SCL"],
    "I2S": ["CK", "MCK", "SD", "WS", "EXTSD"],
    "USART": ["RX", "TX", "CTS", "RTS", "CK"],
    "UART": ["RX", "TX", "CTS", "RTS"],
    "LPUART": ["RX", "TX", "CTS", "RTS"],
    "SPI": ["NSS", "SCK", "MISO", "MOSI"],
    "SDMMC": ["CK", "CMD", "D0", "D1", "D2", "D3"],
    "CAN": ["TX", "RX"],
}

# Only make these AFs available if we actually enable those peripherals (by
# assigning default pins in mpconfigboard.h).
CONDITIONAL_VAR = {
    "I2C": "MICROPY_HW_I2C{num}_SCL",
    "I2S": "MICROPY_HW_I2S{num}",
    "SPI": "MICROPY_HW_SPI{num}_SCK",
    "UART": "MICROPY_HW_UART{num}_TX",
    "LPUART": "MICROPY_HW_LPUART{num}_TX",
    "USART": "MICROPY_HW_UART{num}_TX",
    "SDMMC": "MICROPY_HW_SDMMC{num}_CK",
    "CAN": "MICROPY_HW_CAN{num}_TX",
}

# Contains the result of parsing an cell from af.csv.
PinAf = namedtuple(
    "PinAf",
    [
        "af_idx",  # int, 0-15
        "af_fn",  # e.g. "I2C"
        "af_unit",  # int, e.g. 1 (for I2C1) or None (for OTG_HS_ULPI_CK)
        "af_pin",  # e.g. "SDA"
        "af_supported",  # bool, see table above
        "af_name",  # e.g. "I2C1_SDA"
    ],
)

# Only support ADC1, ADC2, ADC3 for now (e.g. cannot support ADC4 & ADC5 on
# STM32G4).
MIN_ADC_UNIT = 1
MAX_ADC_UNIT = 3


class Stm32Pin(boardgen.Pin):
    def __init__(self, cpu_pin_name):
        super().__init__(cpu_pin_name)

        # Pins ending in "_C" correspond to the analog-only pad of a pair
        # of dual (analog) pads found on H7 MCUs (eg PA0 and PA0_C pair).
        self._analog_only = cpu_pin_name.endswith("_C")
        if self._analog_only:
            cpu_pin_name = cpu_pin_name[:-2]

        # P<port><num> (already verified by validate_cpu_pin_name).
        self._port = cpu_pin_name[1]
        self._pin = int(cpu_pin_name[2:])

        # List of PinAf instances.
        self._afs = []

        # The channel that this pin uses on each of the listed units.
        self._adc_channel = 0
        # e.g. ADC123 -> [1,2,3]
        self._adc_units = []

    # Called for each AF defined in the csv file for this pin. e.g. "SPI2_NSS/I2S2_WS"
    def add_af(self, af_idx, af_name, af):
        if af_idx > 16:
            # The AF csv files should not have more than 16 AFs + 1 ADC list.
            return

        # AF 16 is the ADC, which is special cased below.
        if af_idx == 16:
            if af_name != "ADC":
                raise boardgen.PinGeneratorError(
                    "Invalid AF column name '{:s}' for ADC column with index {:d}.".format(
                        af_name, af_idx
                    )
                )
            return self.add_adc(af)

        if af_name != "AF{:d}".format(af_idx):
            raise boardgen.PinGeneratorError(
                "Invalid AF column name '{:s}' for AF index {:d}.".format(af_name, af_idx)
            )

        # If there is a slash, then the slash separates multiple aliases for
        # the same alternate function.
        for af_name in af.split("/"):
            if not af_name.strip():
                continue
            # This matches <fn><unit>_<pin>, with consideration for:
            #  - fn may contain a digit (e.g. I2C)
            #  - there may be an "ext" after the unit
            m = re.match("([A-Z0-9]+[A-Z])(([0-9]+)(ext)?)?(_(.*))?", af_name)
            if not m:
                raise boardgen.PinGeneratorError(
                    "Invalid af '{:s}' for pin '{:s}'".format(af_name, self.name())
                )
            else:
                af_fn = m.group(1)
                af_unit = int(m.group(3)) if m.group(3) is not None else None
                af_ext = m.group(4) == "ext"
                af_pin = m.group(6)

                # Special case. We change I2S2ext_SD into I2S2_EXTSD so that it parses
                # the same way the other peripherals do.
                if af_ext:
                    af_pin = "EXT" + af_pin

                af_supported = af_fn in SUPPORTED_AF and af_pin in SUPPORTED_AF[af_fn]

                self._afs.append(PinAf(af_idx, af_fn, af_unit, af_pin, af_supported, af_name))

    # ADCs are slash separated "ADC<list of units>_<mode><channel>", where unit=1,2,3,4,5 and mode=IN,INN,INP
    def add_adc(self, adc):
        if not adc.strip():
            return

        # TODO: This needs to be improved to support the case where a pin can
        # be the P for one channel, and the N for a different channel.
        # e.g. "ADC123_INP12/ADC123_INN11".
        for adc_name in adc.split("/"):
            m = re.match("ADC([1-5]+)_(IN[NP]?)([0-9]+)$", adc_name)
            if not m:
                raise boardgen.PinGeneratorError(
                    "Invalid adc '{:s}' for pin '{:s}'".format(adc_name, self.name())
                )
            adc_units = [int(x) for x in m.group(1) if MIN_ADC_UNIT <= int(x) <= MAX_ADC_UNIT]
            adc_mode = m.group(2)
            if adc_mode == "INN":
                # On H7 we have INN/INP, all other parts use IN only. Only use
                # IN or INP channels.
                continue
            adc_channel = int(m.group(3))

            # Pick the entry with the most ADC units, e.g. "ADC1_INP16/ADC12_INN1/ADC12_INP0" --> "ADC12_INN1".
            if len(adc_units) > len(self._adc_units):
                self._adc_units = adc_units
                self._adc_channel = adc_channel

    # STM32-specific behavior, strip the "P" from the start of the name when emitting this pin.
    # e.g. the #define is pin_A11 not pin_PA11. Fortunately we don't have to special case this
    # as there are no places where emit the full "PA11" name.
    def name(self):
        return self._cpu_pin_name[1:]

    # Use the PIN() macro defined in stm32f4xx_prefix.c for defining the pin
    # objects.
    def definition(self):
        # Generate bitfield of supported ADC units where lsb is unit 1 (e.g. [1,3] --> 0b101).
        adc_units_bitfield = (
            " | ".join("PIN_ADC{}".format(unit) for unit in self._adc_units) or "0"
        )

        # PIN(p_port, p_pin, p_af, p_adc_num, p_adc_channel)
        pin_macro = "PIN_ANALOG" if self._analog_only else "PIN"
        return "{:s}({:s}, {:d}, pin_{:s}_af, {:s}, {:d})".format(
            pin_macro, self._port, self._pin, self.name(), adc_units_bitfield, self._adc_channel
        )

    # This will be called at the start of the output (after the prefix). Use
    # it to emit the af objects (via the AF() macro).
    def print_source(self, out_source):
        print(file=out_source)
        print("const pin_af_obj_t pin_{:s}_af[] = {{".format(self.name()), file=out_source)
        for af in self._afs:
            if af.af_fn in CONDITIONAL_VAR:
                print(
                    "    #if defined({:s})".format(
                        CONDITIONAL_VAR[af.af_fn].format(num=af.af_unit)
                    ),
                    file=out_source,
                )
            if af.af_supported:
                print("    ", end="", file=out_source)
            else:
                print("    // ", end="", file=out_source)
            # AF(af_idx, af_fn, af_unit, af_type, af_ptr)
            print(
                "AF({:d}, {:s}, {:d}, {:s}, {:s}{:s}),  // {:s}".format(
                    af.af_idx,
                    af.af_fn,
                    af.af_unit or 0,
                    af.af_pin or "NONE",
                    af.af_fn,
                    "" if af.af_unit is None else str(af.af_unit),
                    af.af_name,
                ),
                file=out_source,
            )
            if af.af_fn in CONDITIONAL_VAR:
                print("    #endif", file=out_source)
        print("};", file=out_source)

    # STM32 cpu names must be "P<port><num>".
    @staticmethod
    def validate_cpu_pin_name(cpu_pin_name):
        boardgen.Pin.validate_cpu_pin_name(cpu_pin_name)

        if not re.match("P[A-K][0-9]+(_C)?$", cpu_pin_name):
            raise boardgen.PinGeneratorError("Invalid cpu pin name '{}'".format(cpu_pin_name))


class Stm32PinGenerator(boardgen.PinGenerator):
    def __init__(self):
        # Use custom pin type above, and also enable the --af-csv argument so
        # that add_af gets called on each pin.
        super().__init__(
            pin_type=Stm32Pin,
            enable_af=True,
        )

    # STM32-specific behavior, we use pin_A0 for the cpu names, but
    # pyb_pin_X11 for the board names.
    def board_name_define_prefix(self):
        return "pyb_"

    # Override the default implementation just to change the default arguments
    # (extra header row, skip first column).
    def parse_af_csv(self, filename):
        return super().parse_af_csv(filename, header_rows=2, pin_col=1, af_col=2)

    # Find which ADCs are used on this chip and on how many pins and the
    # maximum channel number for each.
    def count_adc_pins(self):
        adc_units = defaultdict(lambda: (0, 0))
        for pin in self._pins:  # All pins
            for unit in pin._adc_units:
                num, max_channel = adc_units[unit]
                if pin._available:
                    adc_units[unit] = num + 1, max(max_channel, pin._adc_channel)
        return adc_units.items()

    # Print table of pins for each ADC (indexed by channel).
    def print_adcs(self, out_source):
        for adc_unit, (num_pins, max_channel) in self.count_adc_pins():
            print(file=out_source)
            print(
                "const machine_pin_obj_t * const pin_adc{:d}[{:d}] = {{".format(
                    adc_unit, max_channel + 1
                ),
                file=out_source,
            )
            # Don't include pins that weren't in pins.csv.
            for pin in self.available_pins():
                if pin._hidden:
                    continue
                if adc_unit in pin._adc_units:
                    print(
                        "    [{:d}] = {:s},".format(pin._adc_channel, self._cpu_pin_pointer(pin)),
                        file=out_source,
                    )
            print("};", file=out_source)

    # Print externs for the adc pin tables.
    def print_adc_externs(self, out_source):
        print(file=out_source)
        for adc_unit, (num_pins, max_channel) in self.count_adc_pins():
            print(
                "extern const machine_pin_obj_t * const pin_adc{:d}[{:d}];".format(
                    adc_unit, max_channel + 1
                ),
                file=out_source,
            )

    # Append ADC definitions to the end of the source output.
    def print_source(self, out_source):
        super().print_source(out_source)
        self.print_adcs(out_source)

    # Append ADC externs to the end of the header output, and don't include
    # externs in mboot mode.
    def print_header(self, out_header):
        if self.args.mboot_mode:
            self.print_defines(out_header, cpu=False)
        else:
            super().print_header(out_header)
            self.print_adc_externs(out_header)

    # This is a set of map entries `MP_QSTR_AF<num>_<fn>` -> `GPIO_AF<num>_<fn>`
    # that become part of the locals dict of machine.Pin.
    def print_af_const(self, out_af_const):
        # Extract all unique "AF<num>_<fn>" values.
        names = set()
        for pin in self.available_pins():
            for af in pin._afs:
                if not af.af_supported:
                    continue
                key = (
                    "AF{:d}_{:s}{:d}".format(af.af_idx, af.af_fn, af.af_unit),
                    af.af_fn,
                    af.af_unit,
                )
                names.add(key)

        # Generate the table.
        for key in sorted(names):
            name, af_fn, af_unit = key
            if af_fn in CONDITIONAL_VAR:
                print(
                    "    #if defined({:s})".format(CONDITIONAL_VAR[af_fn].format(num=af_unit)),
                    file=out_af_const,
                )
            print(
                "    {{ MP_ROM_QSTR(MP_QSTR_{:s}), MP_ROM_INT(GPIO_{:s}) }},".format(name, name),
                file=out_af_const,
            )
            if af_fn in CONDITIONAL_VAR:
                print("    #endif", file=out_af_const)

    # Output macros to compile-time match a pin object to its AF. See
    # mp_hal_pin_config_alt_static and mp_hal_pin_config_alt_static_speed in
    # pin_static_af.h.
    def print_af_defs(self, out_af_defs):
        # Get the set of unique "<fn><unit>_<pinfn>" (e.g. I2C1_SDA) and which
        # pins can be used for each.
        af_defs = defaultdict(list)
        for pin in self._pins:
            for af in pin._afs:
                key = af.af_fn, af.af_unit, af.af_pin
                af_defs[key].append((pin, af.af_idx))

        # Emit a macro for each that will match a pin object to the
        # corresponding af index for that pin doing that function.
        for key, pins in af_defs.items():
            af_fn, af_unit, af_pin = key
            print(file=out_af_defs)
            print(
                "#define STATIC_AF_{:s}{:s}_{:s}(pin_obj) ( \\".format(
                    af_fn, "" if af_unit is None else str(af_unit), af_pin or "NULL"
                ),
                file=out_af_defs,
            )
            for pin, af_idx in pins:
                if self.args.mboot_mode:
                    print(
                        "    ((pin_obj) == (pin_{:s})) ? ({:d}) : \\".format(pin.name(), af_idx),
                        file=out_af_defs,
                    )
                else:
                    # Match either "(pin_A11_obj)" (if using pin_A11 or
                    # pyb_pin_X11) or "((pin_A11_obj))" (if going via another
                    # macro e.g. MICROPY_HW_QSPIFLASH_CS).
                    # TODO: Why do we need do do string matching? (i.e. why can't the mboot behavior be used always?).
                    print(
                        '    ((strcmp( #pin_obj , "(&pin_{:s}_obj)") & strcmp( #pin_obj , "((&pin_{:s}_obj))")) == 0) ? ({:d}) : \\'.format(
                            pin.name(), pin.name(), af_idx
                        ),
                        file=out_af_defs,
                    )
            print("    (0xffffffffffffffffULL))", file=out_af_defs)

    # Additional stm32-specific outputs that will be written in
    # generate_extra_files().
    def extra_args(self, parser):
        parser.add_argument("--output-af-const")
        parser.add_argument("--output-af-defs")

        # In mboot mode the af-defs use object rather than string comparison,
        # and we don't include externs in the header file.
        parser.add_argument("--mboot-mode", action="store_true")

    # Called in main() after everything else is done to write additional files.
    def generate_extra_files(self):
        if self.args.output_af_const:
            with open(self.args.output_af_const, "w") as out_af_const:
                self.print_af_const(out_af_const)

        if self.args.output_af_defs:
            with open(self.args.output_af_defs, "w") as out_af_defs:
                self.print_af_defs(out_af_defs)


if __name__ == "__main__":
    Stm32PinGenerator().main()



import socket, ssl, errno, sys, time, select


def test_one(site, opts):
    ai = socket.getaddrinfo(site, 443, socket.AF_INET)
    addr = ai[0][-1]
    print(site)

    # Connect the raw socket
    s = socket.socket(socket.AF_INET)
    s.setblocking(False)
    try:
        s.connect(addr)
        raise OSError(-1, "connect blocks")
    except OSError as e:
        if e.errno != errno.EINPROGRESS:
            raise

    # Create SSLContext.
    ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)

    # CPython compatibility:
    # - disable check_hostname
    # - load default system certificate chain
    # - must wait for socket to be writable before calling wrap_socket
    if sys.implementation.name != "micropython":
        ssl_context.check_hostname = False
        ssl_context.load_default_certs()
        select.select([], [s], [])

    try:
        # Wrap with SSL
        try:
            s = ssl_context.wrap_socket(s, do_handshake_on_connect=False)
        except OSError as e:
            if e.errno != errno.EINPROGRESS:
                raise
        print("wrapped")

        # CPython needs to be told to do the handshake
        if sys.implementation.name != "micropython":
            while True:
                try:
                    s.do_handshake()
                    break
                except ssl.SSLError as err:
                    if err.args[0] == ssl.SSL_ERROR_WANT_READ:
                        select.select([s], [], [])
                    elif err.args[0] == ssl.SSL_ERROR_WANT_WRITE:
                        select.select([], [s], [])
                    else:
                        raise
                time.sleep(0.1)
            # print("shook hands")

        # Write HTTP request
        out = b"GET / HTTP/1.0\r\nHost: %s\r\n\r\n" % bytes(site, "latin")
        while len(out) > 0:
            n = s.write(out)
            if n is None:
                continue
            if n > 0:
                out = out[n:]
            elif n == 0:
                raise OSError(-1, "unexpected EOF in write")
        print("wrote")

        # Read response
        resp = b""
        while True:
            try:
                b = s.read(128)
            except OSError as err:
                if err.errno == 2:  # 2=ssl.SSL_ERROR_WANT_READ:
                    continue
                raise
            if b is None:
                continue
            if len(b) > 0:
                if len(resp) < 1024:
                    resp += b
            elif len(b) == 0:
                break
        print("read")

        if resp[:7] != b"HTTP/1.":
            raise ValueError("response doesn't start with HTTP/1.")
        # print(resp)

    finally:
        s.close()


SITES = [
    "www.github.com",
    "micropython.org",
    "pypi.org",
    {"host": "api.pushbullet.com", "sni": True},
]


def main():
    for site in SITES:
        opts = {}
        if isinstance(site, dict):
            opts = site
            site = opts["host"]
        try:
            test_one(site, opts)
            print(site, "ok")
        except Exception as e:
            print(site, "error", e)
    print("DONE")


main()



# test memoryview accessing maximum values for signed/unsigned elements
try:
    memoryview
except:
    print("SKIP")
    raise SystemExit
try:
    from array import array
except ImportError:
    print("SKIP")
    raise SystemExit

print(list(memoryview(b'\x7f\x80\x81\xff')))
print(list(memoryview(array('b', [0x7f, -0x80]))))
print(list(memoryview(array('B', [0x7f, 0x80, 0x81, 0xff]))))
print(list(memoryview(array('h', [0x7f00, -0x8000]))))
print(list(memoryview(array('H', [0x7f00, 0x8000, 0x8100, 0xffff]))))



# Extended float comparisons


class Foo:
    pass


foo = Foo()

print(foo == 1.0)
print(1.0 == foo)
print(1.0 == Foo)
print(1.0 == [])
print(1.0 == {})

try:
    print(foo < 1.0)
except TypeError:
    print("TypeError")

try:
    print(1.0 < foo)
except TypeError:
    print("TypeError")



# check that interned strings are compared against non-interned strings

di = {"key1": "value"}

# lookup interned string
k = "key1"
print(k in di)

# lookup non-interned string
k2 = "key" + "1"
print(k == k2)
print(k2 in di)

# lookup non-interned string
print("".join(['k', 'e', 'y', '1']) in di)



# Simple test creating an SSL connection and transferring some data
# This test won't run under CPython because CPython doesn't have key/cert

try:
    import binascii, os, socket, ssl
except ImportError:
    print("SKIP")
    raise SystemExit

PORT = 8000

# These are test certificates. See tests/README.md for details.
certfile = "ec_cert.der"
keyfile = "ec_key.der"

try:
    os.stat(certfile)
    os.stat(keyfile)
except OSError:
    print("SKIP")
    raise SystemExit

with open(certfile, "rb") as cf:
    cert = cadata = cf.read()

with open(keyfile, "rb") as kf:
    key = kf.read()


# Server
def instance0():
    multitest.globals(IP=multitest.get_network_ip())
    s = socket.socket()
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    s.bind(socket.getaddrinfo("0.0.0.0", PORT)[0][-1])
    s.listen(1)
    multitest.next()
    s2, _ = s.accept()
    s2 = ssl.wrap_socket(s2, server_side=True, key=key, cert=cert)
    print(s2.read(16))
    s2.write(b"server to client")
    s2.close()
    s.close()


# Client
def instance1():
    multitest.next()
    s = socket.socket()
    s.connect(socket.getaddrinfo(IP, PORT)[0][-1])
    s = ssl.wrap_socket(
        s, cert_reqs=ssl.CERT_REQUIRED, server_hostname="micropython.local", cadata=cadata
    )
    s.write(b"client to server")
    print(s.read(16))
    s.close()



# check a case where rounding was suppressed inappropriately when "f" was
# promoted to "e" for large numbers.
v = 8.888e32
print("%.2f" % v)  # '%.2f' format with e32 becomes '%.2e', expect 8.89e+32.



# Test BLE re-connect after peripheral deepsleep.
# instance0 needs to be a device capable of deepsleep without disconnecting the
# REPL, eg. STM32WB55 Nucleo board with REPL on UART port.

from micropython import const
import time, machine, bluetooth

TIMEOUT_MS = 4000

_IRQ_CENTRAL_CONNECT = const(1)
_IRQ_CENTRAL_DISCONNECT = const(2)
_IRQ_PERIPHERAL_CONNECT = const(7)
_IRQ_PERIPHERAL_DISCONNECT = const(8)

waiting_events = {}


def irq(event, data):
    if event == _IRQ_CENTRAL_CONNECT:
        print("_IRQ_CENTRAL_CONNECT")
        waiting_events[event] = data[0]
    elif event == _IRQ_CENTRAL_DISCONNECT:
        print("_IRQ_CENTRAL_DISCONNECT")
    elif event == _IRQ_PERIPHERAL_CONNECT:
        print("_IRQ_PERIPHERAL_CONNECT")
        waiting_events[event] = data[0]
    elif event == _IRQ_PERIPHERAL_DISCONNECT:
        print("_IRQ_PERIPHERAL_DISCONNECT")

    if event not in waiting_events:
        waiting_events[event] = None


def wait_for_event(event, timeout_ms):
    t0 = time.ticks_ms()
    while time.ticks_diff(time.ticks_ms(), t0) < timeout_ms:
        if event in waiting_events:
            return waiting_events.pop(event)
        machine.idle()
    raise ValueError("Timeout waiting for {}".format(event))


# Acting in peripheral role.
def instance0():
    multitest.globals(BDADDR=ble.config("mac"))
    print("gap_advertise")
    ble.gap_advertise(20_000, b"\x02\x01\x06\x04\tMPY")
    multitest.next()
    # Wait for central to connect, then wait for it to disconnect.
    wait_for_event(_IRQ_CENTRAL_CONNECT, TIMEOUT_MS)
    wait_for_event(_IRQ_CENTRAL_DISCONNECT, TIMEOUT_MS)

    multitest.expect_reboot(resume="instance0_resume", delay_ms=100)
    machine.deepsleep(100)


# Acting in peripheral role.
def instance0_resume():
    # continue here after reboot
    print("restarted")
    print("gap_advertise")
    ble.gap_advertise(20_000, b"\x02\x01\x06\x04\tMPY")
    try:
        # Wait for central to connect, then disconnect it.
        conn_handle = wait_for_event(_IRQ_CENTRAL_CONNECT, TIMEOUT_MS)
        print("gap_disconnect:", ble.gap_disconnect(conn_handle))
        wait_for_event(_IRQ_CENTRAL_DISCONNECT, TIMEOUT_MS)
    finally:
        ble.active(0)


# Acting in central role.
def instance1():
    multitest.next()
    try:
        # Connect to peripheral and then disconnect.
        print("gap_connect")
        ble.gap_connect(*BDADDR)
        conn_handle = wait_for_event(_IRQ_PERIPHERAL_CONNECT, TIMEOUT_MS)
        print("gap_disconnect:", ble.gap_disconnect(conn_handle))
        wait_for_event(_IRQ_PERIPHERAL_DISCONNECT, TIMEOUT_MS)

        # Connect to peripheral and then let the peripheral disconnect us.
        # Extra scan timeout allows for the peripheral to receive the disconnect
        # event and start advertising again.
        print("gap_connect")
        ble.gap_connect(BDADDR[0], BDADDR[1], 5000)
        wait_for_event(_IRQ_PERIPHERAL_CONNECT, TIMEOUT_MS)
        wait_for_event(_IRQ_PERIPHERAL_DISCONNECT, TIMEOUT_MS)
    finally:
        ble.active(0)


ble = bluetooth.BLE()
ble.active(1)
ble.irq(irq)



# test basic float capabilities

# literals
print(0.12)
print(1.0)
print(1.2)
print(0e0)
print(0e0)
print(0e-0)

# float construction
print(float(1.2))
print(float("1.2"))
print(float("+1"))
print(float("1e1"))
print(float("1e+1"))
print(float("1e-1"))
print(float("inf"))
print(float("-inf"))
print(float("INF"))
print(float("infinity"))
print(float("INFINITY"))
print(float("nan"))
print(float("-nan"))
print(float("NaN"))
try:
    float("")
except ValueError:
    print("ValueError")
try:
    float("1e+")
except ValueError:
    print("ValueError")
try:
    float("1z")
except ValueError:
    print("ValueError")

# construct from something with the buffer protocol
print(float(b"1.2"))
print(float(bytearray(b"3.4")))

# unary operators
print(bool(0.0))
print(bool(1.2))
print(+(1.2))
print(-(1.2))

# division of integers
x = 1 / 2
print(x)

# /= operator
a = 1
a /= 2
print(a)

# floor division
print(1.0 // 2)
print(2.0 // 2)

# comparison
print(1.2 <= 3.4)
print(1.2 <= -3.4)
print(1.2 >= 3.4)
print(1.2 >= -3.4)
print(0.0 == False, 1.0 == True)
print(False == 0.0, True == 1.0)

# comparison of nan is special
nan = float("nan")
print(nan == 1.2)
print(nan == nan)

try:
    1.0 / 0
except ZeroDivisionError:
    print("ZeroDivisionError")

try:
    1.0 // 0
except ZeroDivisionError:
    print("ZeroDivisionError")

try:
    1.2 % 0
except ZeroDivisionError:
    print("ZeroDivisionError")

try:
    0.0**-1
except ZeroDivisionError:
    print("ZeroDivisionError")

# unsupported unary ops

try:
    ~1.2
except TypeError:
    print("TypeError")

try:
    1.2 in 3.4
except TypeError:
    print("TypeError")

# small int on LHS, float on RHS, unsupported op
try:
    print(1 | 1.0)
except TypeError:
    print("TypeError")

# can't convert list to float
try:
    float([])
except TypeError:
    print("TypeError")

# test constant float with more than 255 chars
x = 1.84728699436059052516398251149631771898472869943605905251639825114963177189847286994360590525163982511496317718984728699436059052516398251149631771898472869943605905251639825114963177189847286994360590525163982511496317718984728699436059052516398251149631771898472869943605905251639825114963177189
print("%.5f" % x)



from umqtt.simple import MQTTClient

# Test reception e.g. with:
# mosquitto_sub -t foo_topic


def main(server="localhost"):
    c = MQTTClient("umqtt_client", server)
    c.connect()
    c.publish(b"foo_topic", b"hello")
    c.disconnect()


if __name__ == "__main__":
    main()



import glob
import os
import os.path
import shutil
import sys
import unittest

from test.support import run_unittest, TESTFN, skip_unless_symlink, can_symlink, create_empty_file


class GlobTests(unittest.TestCase):
    def norm(self, *parts):
        return os.path.normpath(os.path.join(self.tempdir, *parts))

    def mktemp(self, *parts):
        filename = self.norm(*parts)
        base, file = os.path.split(filename)
        if not os.path.exists(base):
            os.makedirs(base)
        create_empty_file(filename)

    def setUp(self):
        self.tempdir = TESTFN + "_dir"
        self.mktemp("a", "D")
        self.mktemp("aab", "F")
        self.mktemp(".aa", "G")
        self.mktemp(".bb", "H")
        self.mktemp("aaa", "zzzF")
        self.mktemp("ZZZ")
        self.mktemp("a", "bcd", "EF")
        self.mktemp("a", "bcd", "efg", "ha")
        if can_symlink():
            os.symlink(self.norm("broken"), self.norm("sym1"))
            os.symlink("broken", self.norm("sym2"))
            os.symlink(os.path.join("a", "bcd"), self.norm("sym3"))

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def glob(self, *parts):
        if len(parts) == 1:
            pattern = parts[0]
        else:
            pattern = os.path.join(*parts)
        p = os.path.join(self.tempdir, pattern)
        res = glob.glob(p)
        self.assertEqual(list(glob.iglob(p)), res)
        bres = [os.fsencode(x) for x in res]
        # Bytes globbing is not support on MicroPython
        #        self.assertEqual(glob.glob(os.fsencode(p)), bres)
        #        self.assertEqual(list(glob.iglob(os.fsencode(p))), bres)
        return res

    def assertSequencesEqual_noorder(self, l1, l2):
        l1 = list(l1)
        l2 = list(l2)
        self.assertEqual(set(l1), set(l2))
        self.assertEqual(sorted(l1), sorted(l2))

    def test_glob_literal(self):
        eq = self.assertSequencesEqual_noorder
        eq(self.glob("a"), [self.norm("a")])
        eq(self.glob("a", "D"), [self.norm("a", "D")])
        eq(self.glob("aab"), [self.norm("aab")])
        eq(self.glob("zymurgy"), [])

        res = glob.glob("*")
        self.assertEqual({type(r) for r in res}, {str})
        res = glob.glob(os.path.join(os.curdir, "*"))
        self.assertEqual({type(r) for r in res}, {str})

    #        res = glob.glob(b'*')
    #        self.assertEqual({type(r) for r in res}, {bytes})
    #        res = glob.glob(os.path.join(os.fsencode(os.curdir), b'*'))
    #        self.assertEqual({type(r) for r in res}, {bytes})

    def test_glob_one_directory(self):
        eq = self.assertSequencesEqual_noorder
        eq(self.glob("a*"), map(self.norm, ["a", "aab", "aaa"]))
        eq(self.glob("*a"), map(self.norm, ["a", "aaa"]))
        eq(self.glob(".*"), map(self.norm, [".aa", ".bb"]))
        eq(self.glob("?aa"), map(self.norm, ["aaa"]))
        eq(self.glob("aa?"), map(self.norm, ["aaa", "aab"]))
        eq(self.glob("aa[ab]"), map(self.norm, ["aaa", "aab"]))
        eq(self.glob("*q"), [])

    def test_glob_nested_directory(self):
        eq = self.assertSequencesEqual_noorder
        if os.path.normcase("abCD") == "abCD":
            # case-sensitive filesystem
            eq(self.glob("a", "bcd", "E*"), [self.norm("a", "bcd", "EF")])
        else:
            # case insensitive filesystem
            eq(
                self.glob("a", "bcd", "E*"),
                [self.norm("a", "bcd", "EF"), self.norm("a", "bcd", "efg")],
            )
        eq(self.glob("a", "bcd", "*g"), [self.norm("a", "bcd", "efg")])

    def test_glob_directory_names(self):
        eq = self.assertSequencesEqual_noorder
        eq(self.glob("*", "D"), [self.norm("a", "D")])
        eq(self.glob("*", "*a"), [])
        eq(self.glob("a", "*", "*", "*a"), [self.norm("a", "bcd", "efg", "ha")])
        eq(self.glob("?a?", "*F"), [self.norm("aaa", "zzzF"), self.norm("aab", "F")])

    def test_glob_directory_with_trailing_slash(self):
        # Patterns ending with a slash shouldn't match non-dirs
        res = glob.glob(self.norm("Z*Z") + os.sep)
        self.assertEqual(res, [])
        res = glob.glob(self.norm("ZZZ") + os.sep)
        self.assertEqual(res, [])
        # When there is a wildcard pattern which ends with os.sep, glob()
        # doesn't blow up.
        res = glob.glob(self.norm("aa*") + os.sep)
        self.assertEqual(len(res), 2)
        # either of these results is reasonable
        self.assertIn(
            set(res),
            [
                {self.norm("aaa"), self.norm("aab")},
                {self.norm("aaa") + os.sep, self.norm("aab") + os.sep},
            ],
        )

    @unittest.skip("unsupported on MicroPython")
    def test_glob_bytes_directory_with_trailing_slash(self):
        # Same as test_glob_directory_with_trailing_slash, but with a
        # bytes argument.
        res = glob.glob(os.fsencode(self.norm("Z*Z") + os.sep))
        self.assertEqual(res, [])
        res = glob.glob(os.fsencode(self.norm("ZZZ") + os.sep))
        self.assertEqual(res, [])
        res = glob.glob(os.fsencode(self.norm("aa*") + os.sep))
        self.assertEqual(len(res), 2)
        # either of these results is reasonable
        self.assertIn(
            set(res),
            [
                {os.fsencode(self.norm("aaa")), os.fsencode(self.norm("aab"))},
                {os.fsencode(self.norm("aaa") + os.sep), os.fsencode(self.norm("aab") + os.sep)},
            ],
        )

    @skip_unless_symlink
    def test_glob_symlinks(self):
        eq = self.assertSequencesEqual_noorder
        eq(self.glob("sym3"), [self.norm("sym3")])
        eq(self.glob("sym3", "*"), [self.norm("sym3", "EF"), self.norm("sym3", "efg")])
        self.assertIn(
            self.glob("sym3" + os.sep), [[self.norm("sym3")], [self.norm("sym3") + os.sep]]
        )
        eq(
            self.glob("*", "*F"),
            [self.norm("aaa", "zzzF"), self.norm("aab", "F"), self.norm("sym3", "EF")],
        )

    @skip_unless_symlink
    def test_glob_broken_symlinks(self):
        eq = self.assertSequencesEqual_noorder
        eq(self.glob("sym*"), [self.norm("sym1"), self.norm("sym2"), self.norm("sym3")])
        eq(self.glob("sym1"), [self.norm("sym1")])
        eq(self.glob("sym2"), [self.norm("sym2")])

    @unittest.skipUnless(sys.platform == "win32", "Win32 specific test")
    def test_glob_magic_in_drive(self):
        eq = self.assertSequencesEqual_noorder
        eq(glob.glob("*:"), [])
        eq(glob.glob(b"*:"), [])
        eq(glob.glob("?:"), [])
        eq(glob.glob(b"?:"), [])
        eq(glob.glob("\\\\?\\c:\\"), ["\\\\?\\c:\\"])
        eq(glob.glob(b"\\\\?\\c:\\"), [b"\\\\?\\c:\\"])
        eq(glob.glob("\\\\*\\*\\"), [])
        eq(glob.glob(b"\\\\*\\*\\"), [])


def test_main():
    run_unittest(GlobTests)


if __name__ == "__main__":
    test_main()



# test loading from bytes and bytearray (introduced in Python 3.6)

try:
    import json
except ImportError:
    print("SKIP")
    raise SystemExit

print(json.loads(b"[1,2]"))
print(json.loads(bytearray(b"[null]")))



#!/usr/bin/env python3
"""
WebAssembly build tools for the NEAR Python contract compiler.
"""

import shutil
import sys
from pathlib import Path
from typing import Set

from .abi import inject_abi
from .analyzer import analyze_contract, find_imports
from .exports import inject_contract_exports
from .manifest import prepare_build_files
from .metadata import inject_metadata_function
from .utils import console, run_command_with_progress, with_progress


@with_progress("Building MicroPython cross-compiler")
def build_mpy_cross(
    mpy_cross_dir: Path,
    build_dir: Path,
    rebuild: bool = False,
    progress=None,
    task_id=None,
) -> Path:
    """
    Build the MicroPython cross-compiler.

    Args:
        mpy_cross_dir: Path to the mpy-cross directory
        build_dir: Path to the build directory
        rebuild: Whether to force a rebuild
        progress: Progress instance
        task_id: Task ID in the progress bar

    Returns:
        Path to the mpy-cross executable
    """
    mpy_cross_build_dir = build_dir / "mpy-cross"
    mpy_cross_exe = mpy_cross_build_dir / "mpy-cross"

    if mpy_cross_exe.exists() and not rebuild:
        # Skip showing the progress bar completely for cached builds
        if progress:
            progress.stop()
        console.print("[cyan]Using existing MicroPython cross-compiler[/]")
        return mpy_cross_exe

    # Only reach here if we need to build
    mpy_cross_build_dir.mkdir(exist_ok=True)

    if not run_command_with_progress(
        ["make", "-C", str(mpy_cross_dir), f"BUILD={mpy_cross_build_dir}"],
        track_task_id=task_id,
        progress=progress,
        description="Building MicroPython cross-compiler",
    ):
        console.print("[red]Failed to build MicroPython cross-compiler")
        sys.exit(1)

    return mpy_cross_exe


@with_progress("Compiling WebAssembly contract")
def build_wasm(
    mpy_port_dir: Path,
    build_dir: Path,
    mpy_cross_exe: Path,
    manifest_path: Path,
    wrappers_path: Path,
    exports: Set[str],
    output_path: Path,
    progress=None,
    task_id=None,
) -> bool:
    """
    Build the WebAssembly contract.

    Args:
        mpy_port_dir: Path to the MicroPython port directory
        build_dir: Path to the build directory
        mpy_cross_exe: Path to the mpy-cross executable
        manifest_path: Path to the manifest file
        wrappers_path: Path to the wrappers file
        exports: Set of exported function names
        output_path: Path where the output WASM should be written
        progress: Progress instance
        task_id: Task ID in the progress bar

    Returns:
        True if compilation succeeded, False if it failed
    """
    # Remove any existing frozen content file to force regeneration
    frozen_content_path = build_dir / "frozen_content.c"
    if frozen_content_path.exists():
        frozen_content_path.unlink()

    # Build command
    build_cmd = [
        "make",
        "-C",
        str(mpy_port_dir),
        f"BUILD={build_dir}",
        f"MICROPY_MPYCROSS={mpy_cross_exe}",
        f"MICROPY_MPYCROSS_DEPENDENCY={mpy_cross_exe}",
        f"FROZEN_MANIFEST={manifest_path}",
        f"SRC_C_GENERATED={wrappers_path}",
        f"EXPORTED_FUNCTIONS={','.join(['_' + e for e in sorted(exports)])}",
        f"OUTPUT_WASM={output_path}",
    ]

    return run_command_with_progress(
        build_cmd,
        track_task_id=task_id,
        progress=progress,
        description="Compiling WebAssembly contract",
    )


def compile_contract(
    contract_path: Path,
    output_path: Path,
    venv_path: Path,
    assets_dir: Path,
    rebuild: bool = False,
    single_file: bool = False,
) -> bool:
    """
    Compile a NEAR contract to WebAssembly with progress display.

    Args:
        contract_path: Path to the contract file
        output_path: Path where the output WASM should be written
        venv_path: Path to the virtual environment
        assets_dir: Path to the assets directory
        rebuild: Whether to force a clean rebuild
        single_file: Whether to skip local module discovery and compile only the specified file

    Returns:
        True if compilation succeeded, False if it failed
    """
    # Setup paths
    build_dir = contract_path.parent / "build"
    mpy_cross_dir = assets_dir / "micropython" / "mpy-cross"
    mpy_port_dir = assets_dir / "micropython" / "ports" / "webassembly-near"

    # Ensure build directory exists
    if rebuild and build_dir.exists():
        shutil.rmtree(build_dir)
    build_dir.mkdir(exist_ok=True)

    # Show a header for the compilation
    console.print(f"[bold cyan]Compiling NEAR Contract:[/] [yellow]{contract_path}[/]")
    if single_file:
        console.print("[cyan]Single file mode: skipping local module discovery[/]")

    # Inject exports for class-based contracts
    contract_with_exports = inject_contract_exports(contract_path)

    # Inject ABI
    contract_with_abi = inject_abi(contract_with_exports)

    # Inject metadata if needed
    contract_with_metadata = inject_metadata_function(contract_with_abi)

    # Use the potentially modified contract for compilation
    # We'll analyze the original contract for exports and imports first to avoid confusion
    exports, imports = analyze_contract(contract_path)

    # Add any additional imports needed for metadata
    if contract_with_metadata != contract_path:
        metadata_imports = find_imports(contract_with_metadata)
        imports = imports.union(metadata_imports)

    # Generate build files
    manifest_file, wrappers_path = prepare_build_files(
        contract_with_metadata, imports, exports, venv_path, build_dir, single_file
    )

    # Build MicroPython cross-compiler if needed
    mpy_cross_exe = build_mpy_cross(mpy_cross_dir, build_dir, rebuild)

    # Build the WASM contract
    if not build_wasm(
        mpy_port_dir,
        build_dir,
        mpy_cross_exe,
        manifest_file,
        wrappers_path,
        exports,
        output_path,
    ):
        console.print("[red]Failed to build WebAssembly contract")
        return False

    # Clean up temporary file if we created one
    if contract_with_metadata != contract_path and contract_with_metadata.exists():
        contract_with_metadata.unlink()

    if contract_with_abi != contract_path and contract_with_abi.exists():
        contract_with_abi.unlink()

    if contract_with_exports != contract_path and contract_with_exports.exists():
        contract_with_exports.unlink()

    # Verify the output file exists
    if not output_path.exists():
        console.print(f"[red]Error: Output file {output_path} was not created")
        return False

    # Show success message with file size
    size_kb = output_path.stat().st_size / 1024
    console.print(
        f"[bold green]Successfully compiled contract:[/] [cyan]{output_path}[/] [yellow]({size_kb:.1f} KB)[/]"
    )
    return True



# Simple test of a ESPnow server and client transferring data.
# This test works with ESP32 or ESP8266 as server or client.

try:
    import network
    import espnow
except ImportError:
    print("SKIP")
    raise SystemExit

# Set read timeout to 5 seconds
timeout_ms = 5000
default_pmk = b"MicroPyth0nRules"


def init(sta_active=True, ap_active=False):
    wlans = [network.WLAN(i) for i in [network.STA_IF, network.AP_IF]]
    e = espnow.ESPNow()
    e.active(True)
    e.set_pmk(default_pmk)
    wlans[0].active(sta_active)
    wlans[1].active(ap_active)
    wlans[0].disconnect()  # Force esp8266 STA interface to disconnect from AP
    e.set_pmk(default_pmk)
    return e


# Server
def instance0():
    e = init(True, False)
    multitest.globals(PEERS=[network.WLAN(i).config("mac") for i in (0, 1)])
    multitest.next()
    peer, msg1 = e.recv(timeout_ms)
    if msg1 is None:
        print("e.recv({timeout_ms}): Timeout waiting for message.")
        e.active(False)
        return
    print(bytes(msg1))
    msg2 = b"server to client"
    e.add_peer(peer)
    e.send(peer, msg2)
    print(bytes(msg2))
    e.active(False)


# Client
def instance1():
    e = init(True, False)
    multitest.next()
    peer = PEERS[0]
    e.add_peer(peer)
    msg1 = b"client to server"
    e.send(peer, msg1)
    print(bytes(msg1))
    peer2, msg2 = e.recv(timeout_ms)
    print(bytes(msg2))
    e.active(False)



import _thread


class Thread:
    def __init__(self, group=None, target=None, name=None, args=(), kwargs=None):
        self.target = target
        self.args = args
        self.kwargs = {} if kwargs is None else kwargs

    def start(self):
        _thread.start_new_thread(self.run, ())

    def run(self):
        self.target(*self.args, **self.kwargs)



# MicroPython LoRa modem driver base class
# MIT license; Copyright (c) 2023 Angus Gratton
#
# LoRa is a registered trademark or service mark of Semtech Corporation or its affiliates.
import time
from micropython import const, schedule

# Set to True to get some additional printed debug output.
_DEBUG = const(False)


def _clamp(v, vmin, vmax):
    # Small utility function to clamp a value 'v' between 'vmin' and 'vmax', inclusive.
    return min(max(vmin, v), vmax)


def _flag(value, condition):
    # Small utility function for returning a bit 'value' or not, based on a
    # boolean condition. Can help make expressions to build register values more
    # readable.
    #
    # Note that for value==1, can also rely on int(bool(x)) with one or both
    # conversions being implicit, as int(True)==1 and int(False)==0
    #
    # There is also (condition and value) but this is (IMO) confusing to read.
    return value if condition else 0


class ConfigError(ValueError):
    # Raise if there is an error in lora_cfg, saves some duplicated strings
    def __init__(self, field):
        super().__init__("Invalid lora_cfg {}".format(field))


class BaseModem:
    def __init__(self, ant_sw):
        self._ant_sw = ant_sw
        self._irq_callback = None

        # Common configuration settings that need to be tracked by all modem drivers.
        #
        # Where modem hardware sets different values after reset, the driver should
        # set them back to these defaults (if not provided by the user), so that
        # behaviour remains consistent between different modems using the same driver.
        self._rf_freq_hz = 0  # Needs to be set via configure()
        self._sf = 7  # Spreading factor
        self._bw_hz = 125000  # Reset value
        self._coding_rate = 5
        self._crc_en = True  # use packet CRCs
        self._implicit_header = False  # implict vs explicit header mode
        self._preamble_len = 12
        self._coding_rate = 5

        # CRC error counter
        self.crc_errors = 0
        self.rx_crc_error = False

        # Current state of the modem

        # _rx holds radio recv state:
        #
        # - False if the radio is not receiving
        # - True if the radio is continuously receiving, or performing a single receive with
        #   no timeout.
        # - An int if there is a timeout set, in which case it is the is the receive deadline
        #   (as a time.ticks_ms() timestamp).
        #
        # Note that self._rx can be not-False even when the radio hardware is not actually
        # receiving, if self._tx is True (send always pauses recv.)
        self._rx = False

        # _rx_continuous is True if the modem is in continuous receive mode
        # (this value is only valid when self._rx is also True).
        self._rx_continuous = False

        # This argument is stored from the parameter of the same name, as set in
        # the last call to start_recv()
        self._rx_length = None

        # _tx holds radio send state and is simpler, True means sending and
        # False means not sending.
        self._tx = False

        # timestamp (as time.ticks_ms() result) of last IRQ event
        self._last_irq = None

        # values are:
        # - lora_cfg["invert_iq_rx"]
        # - lora_cfg["invert_iq_tx"]
        # - Current modem Invert setting
        self._invert_iq = [False, False, False]

        # This hook exists to allow the SyncModem & AsyncModem "mixin-like"
        # classes to have some of their own state, without needing to manage the
        # fuss of multiple constructor paths.
        try:
            self._after_init()
        except AttributeError:
            # If this exception happens here then one of the modem classes without a SyncModem or AsyncModem "mixin-like" class
            # has been instantiated.
            raise NotImplementedError(
                "Don't instantiate this class directly, "
                "instantiate a class from the 'lora' package"
            )

    def standby(self):
        # Put the modem into standby. Can be used to cancel a continuous recv,
        # or cancel a send before it completes.
        #
        # Calls the private function which actually sets the mode to standby, and then
        # clears all the driver's state flags.
        #
        # Note this is also called before going to sleep(), to save on duplicated code.
        self._standby()
        self._rx = False
        self._tx = False
        self._last_irq = None
        if self._ant_sw:
            self._ant_sw.idle()
        self._radio_isr(None)  # "soft ISR"

    def _get_t_sym_us(self):
        # Return length of a symbol in microseconds
        return 1000_000 * (1 << self._sf) // self._bw_hz

    def _get_ldr_en(self):
        # Return true if Low Data Rate should be enabled
        #
        # The calculation in get_n_symbols_x4() relies on this being the same logic applied
        # in the modem configuration routines.
        return self._get_t_sym_us() >= 16000

    def _get_pa_ramp_val(self, lora_cfg, supported):
        # Return the PA ramp register index from the list of supported PA ramp
        # values. If the requested ramp time is supported by the modem, round up
        # to the next supported value.
        #
        # 'supported' is the list of supported ramp times, must be sorted
        # already.
        us = int(lora_cfg["pa_ramp_us"])

        # Find the index of the lowest supported ramp time that is longer or the
        # same value as 'us'
        for i, v in enumerate(supported):
            if v >= us:
                return i
        # The request ramp time is longer than all this modem's supported ramp times
        raise ConfigError("pa_ramp_us")

    def _symbol_offsets(self):
        # Called from get_time_on_air_us().
        #
        # This function provides a way to implement the different SF5 and SF6 in SX126x,
        # by returning two offsets: one for the overall number of symbols, and one for the
        # number of bits used to calculate the symbol length of the payload.
        return (0, 0)

    def get_n_symbols_x4(self, payload_len):
        # Get the number of symbols in a packet (Time-on-Air) for the current
        # configured modem settings and the provided payload length in bytes.
        #
        # Result is in units of "symbols times 4" as there is a fractional term
        # in the equation, and we want to limit ourselves to integer arithmetic.
        #
        # References are:
        # - SX1261/2 DS 6.1.4 "LoRa Time-on-Air"
        # - SX1276 DS 4.1.1 "Time on air"
        #
        # Note the two datasheets give the same information in different
        # ways. SX1261/62 DS is (IMO) clearer, so this function is based on that
        # formula. The result is equivalent to the datasheet value "Nsymbol",
        # times 4.
        #
        # Note also there are unit tests for this function in tests/test_time_on_air.py,
        # and that it's been optimised a bit for code size (with impact on readability)

        # Account for a minor difference between SX126x and SX127x: they have
        # incompatible SF 5 & 6 modes.
        #
        # In SX126x when using SF5 or SF6, we apply an offset of +2 symbols to
        # the overall preamble symbol count (s_o), and an offset of -8 to the
        # payload bit length (b_o).
        s_o, b_o = self._symbol_offsets()

        # calculate the bit length of the payload
        #
        # This is the part inside the max(...,0) in the datasheet
        bits = (
            # payload_bytes
            8 * payload_len
            # N_bit_crc
            + (16 if self._crc_en else 0)
            # (4 * SF)
            - (4 * self._sf)
            # +8 for most modes, except SF5/6 on SX126x where b_o == -8 so these two cancel out
            + 8
            + b_o
            # N_symbol_header
            + (0 if self._implicit_header else 20)
        )
        bits = max(bits, 0)

        # "Bits per symbol" denominator is either (4 * SF) or (4 * (SF -2))
        # depending on Low Data Rate Optimization
        bps = (self._sf - (2 * self._get_ldr_en())) * 4

        return (
            # Fixed preamble portion (4.25), times 4
            17
            # Remainder of equation is an integer number of symbols, times 4
            + 4
            * (
                # configured preamble length
                self._preamble_len
                +
                # optional extra preamble symbols (4.25+2=6.25 for SX1262 SF5,SF6)
                s_o
                +
                # 8 symbol constant overhead
                8
                +
                # Payload symbol length
                # (this is the term "ceil(bits / 4 * SF) * (CR + 4)" in the datasheet
                ((bits + bps - 1) // bps) * self._coding_rate
            )
        )

    def get_time_on_air_us(self, payload_len):
        # Return the "Time on Air" in microseconds for a particular
        # payload length and the current configured modem settings.
        return self._get_t_sym_us() * self.get_n_symbols_x4(payload_len) // 4

    # Modem ISR routines
    #
    # ISR implementation is relatively simple, just exists to signal an optional
    # callback, record a timestamp, and wake up the hardware if
    # needed. Application code is expected to call poll_send() or
    # poll_recv() as applicable in order to confirm the modem state.
    #
    # This is a MP hard irq in some configurations.
    def _radio_isr(self, _):
        self._last_irq = time.ticks_ms()
        if self._irq_callback:
            self._irq_callback()
        if _DEBUG:
            print("_radio_isr")

    def irq_triggered(self):
        # Returns True if the ISR has executed since the last time a send or a receive
        # started
        return self._last_irq is not None

    def set_irq_callback(self, callback):
        # Set a function to be called from the radio ISR
        #
        # This is used by the AsyncModem implementation, but can be called in
        # other circumstances to implement custom ISR logic.
        #
        # Note that callback may be called in hard ISR context.
        self._irq_callback = callback

    def _get_last_irq(self):
        # Return the _last_irq timestamp if set by an ISR, or the
        # current time.time_ms() timestamp otherwise.
        if self._last_irq is None:
            return time.ticks_ms()
        return self._last_irq

    # Common parts of receive API

    def start_recv(self, timeout_ms=None, continuous=False, rx_length=0xFF):
        # Start receiving.
        #
        # Part of common low-level modem API, see README.md for usage.
        if continuous and timeout_ms is not None:
            raise ValueError  # these two options are mutually exclusive

        if timeout_ms is not None:
            self._rx = time.ticks_add(time.ticks_ms(), timeout_ms)
        else:
            self._rx = True

        self._rx_continuous = continuous
        self._rx_length = rx_length

        if self._ant_sw and not self._tx:
            # this is guarded on 'not self._tx' as the subclass will not immediately
            # start receiving if a send is in progress.
            self._ant_sw.rx()

    def poll_recv(self, rx_packet=None):
        # Should be called while a receive is in progress:
        #
        # Part of common low-level modem API, see README.md for usage.
        #
        # This function may alter the state of the modem - it will clear
        # RX interrupts, it may read out a packet from the FIFO, and it
        # may resume receiving if the modem has gone to standby but receive
        # should resume.

        if self._rx is False:
            # Not actually receiving...
            return False

        if self._tx:
            # Actually sending, this has to complete before we
            # resume receiving, but we'll indicate that we are still receiving.
            #
            # (It's not harmful to fall through here and check flags anyhow, but
            # it is a little wasteful if an interrupt has just triggered
            # poll_send() as well.)
            return True

        packet = None

        flags = self._get_irq()

        if _DEBUG and flags:
            print("RX flags {:#x}".format(flags))
        if flags & self._IRQ_RX_COMPLETE:
            # There is a small potential for race conditions here in continuous
            # RX mode. If packets are received rapidly and the call to this
            # function delayed, then a ValidHeader interrupt (for example) might
            # have already set for a second packet which is being received now,
            # and clearing it will mark the second packet as invalid.
            #
            # However it's necessary in continuous mode as interrupt flags don't
            # self-clear in the modem otherwise (for example, if a CRC error IRQ
            # bit sets then it stays set on the next packet, even if that packet
            # has a valid CRC.)
            self._clear_irq(flags)
            ok = self._rx_flags_success(flags)
            if not ok:
                # If a non-valid receive happened, increment the CRC error counter
                self.crc_errors += 1
            if ok or self.rx_crc_error:
                # Successfully received a valid packet (or configured to return all packets)
                packet = self._read_packet(rx_packet, flags)
                if not self._rx_continuous:
                    # Done receiving now
                    self._end_recv()

        # _check_recv() will return True if a receive is ongoing and hasn't timed out,
        # and also manages resuming any modem receive if needed
        #
        # We need to always call check_recv(), but if we received a packet then this is what
        # we should return to the caller.
        res = self._check_recv()
        return packet or res

    def _end_recv(self):
        # Utility function to clear the receive state
        self._rx = False
        if self._ant_sw:
            self._ant_sw.idle()

    def _check_recv(self):
        # Internal function to automatically call start_recv()
        # again if a receive has been interrupted and the host
        # needs to start it again.
        #
        # Return True if modem is still receiving (or sending, but will
        # resume receiving after send finishes).

        if not self._rx:
            return False  # Not receiving, nothing to do

        if not self.is_idle():
            return True  # Radio is already sending or receiving

        rx = self._rx

        timeout_ms = None
        if isinstance(rx, int):  # timeout is set
            timeout_ms = time.ticks_diff(rx, time.ticks_ms())
            if timeout_ms <= 0:
                # Timed out in software, nothing to resume
                self._end_recv()
                if _DEBUG:
                    print("Timed out in software timeout_ms={}".format(timeout_ms))
                schedule(
                    self._radio_isr, None
                )  # "soft irq" to unblock anything waiting on the interrupt event
                return False

        if _DEBUG:
            print(
                "Resuming receive timeout_ms={} continuous={} rx_length={}".format(
                    timeout_ms, self._rx_continuous, self._rx_length
                )
            )

        self.start_recv(timeout_ms, self._rx_continuous, self._rx_length)

        # restore the previous version of _rx so ticks_ms deadline can't
        # slowly creep forward each time this happens
        self._rx = rx

        return True

    # Common parts of send API

    def poll_send(self):
        # Check the ongoing send state.
        #
        # Returns one of:
        #
        # - True if a send is ongoing and the caller
        #   should call again.
        # - False if no send is ongoing.
        # - An int value exactly one time per transmission, the first time
        #   poll_send() is called after a send ends. In this case it
        #   is the time.ticks_ms() timestamp of the time that the send completed.
        #
        # Note this function only returns an int value one time (the first time it
        # is called after send completes).
        #
        # Part of common low-level modem API, see README.md for usage.
        if not self._tx:
            return False

        ticks_ms = self._get_last_irq()

        if not (self._get_irq() & self._IRQ_TX_COMPLETE):
            # Not done. If the host and modem get out
            # of sync here, or the caller doesn't follow the sequence of
            # send operations exactly, then can end up in a situation here
            # where the modem has stopped sending and has gone to Standby,
            # so _IRQ_TX_DONE is never set.
            #
            # For now, leaving this for the caller to do correctly. But if it becomes an issue then
            # we can call _get_mode() here as well and check the modem is still in a TX mode.
            return True

        self._clear_irq()

        self._tx = False

        if self._ant_sw:
            self._ant_sw.idle()

        # The modem just finished sending, so start receiving again if needed
        self._check_recv()

        return ticks_ms


class RxPacket(bytearray):
    # A class to hold a packet received from a LoRa modem.
    #
    # The base class is bytearray, which represents the packet payload,
    # allowing RxPacket objects to be passed anywhere that bytearrays are
    # accepted.
    #
    # Some additional properties are set on the object to store metadata about
    # the received packet.
    def __init__(self, payload, ticks_ms=None, snr=None, rssi=None, valid_crc=True):
        super().__init__(payload)
        self.ticks_ms = ticks_ms
        self.snr = snr
        self.rssi = rssi
        self.valid_crc = valid_crc

    def __repr__(self):
        return "{}({}, {}, {}, {}, {})".format(
            "RxPacket",
            repr(
                bytes(self)
            ),  # This is a bit wasteful, but gets us b'XYZ' rather than "bytearray(b'XYZ')"
            self.ticks_ms,
            self.snr,
            self.rssi,
            self.valid_crc,
        )



metadata(version="2.5.0")

# This is a collection package that gets all hash functions. To save code and
# memory size, prefer to install just the algorithm you need.
require("hashlib-sha224")
require("hashlib-sha256")
require("hashlib-sha384")
require("hashlib-sha512")



# Test concurrency between filesystem access and BLE host. This is
# particularly relevant on STM32WB where the second core is stalled while
# flash operations are in progress.

from micropython import const
import time, machine, bluetooth, os

TIMEOUT_MS = 10000

LOG_PATH_INSTANCE0 = "stress_log_filesystem_0.log"
LOG_PATH_INSTANCE1 = "stress_log_filesystem_1.log"

_IRQ_CENTRAL_CONNECT = const(1)
_IRQ_CENTRAL_DISCONNECT = const(2)
_IRQ_GATTS_WRITE = const(3)
_IRQ_GATTS_READ_REQUEST = const(4)
_IRQ_PERIPHERAL_CONNECT = const(7)
_IRQ_PERIPHERAL_DISCONNECT = const(8)
_IRQ_GATTC_SERVICE_RESULT = const(9)
_IRQ_GATTC_SERVICE_DONE = const(10)
_IRQ_GATTC_CHARACTERISTIC_RESULT = const(11)
_IRQ_GATTC_CHARACTERISTIC_DONE = const(12)
_IRQ_GATTC_READ_RESULT = const(15)
_IRQ_GATTC_READ_DONE = const(16)
_IRQ_GATTC_WRITE_DONE = const(17)

SERVICE_UUID = bluetooth.UUID("A5A5A5A5-FFFF-9999-1111-5A5A5A5A5A5A")
CHAR_UUID = bluetooth.UUID("00000000-1111-2222-3333-444444444444")
CHAR = (
    CHAR_UUID,
    bluetooth.FLAG_READ | bluetooth.FLAG_WRITE | bluetooth.FLAG_NOTIFY | bluetooth.FLAG_INDICATE,
)
SERVICE = (
    SERVICE_UUID,
    (CHAR,),
)
SERVICES = (SERVICE,)


waiting_events = {}
log_file = None


def write_log(*args):
    if log_file:
        print(*args, file=log_file)
        log_file.flush()


last_file_write = 0


def periodic_log_write():
    global last_file_write
    t = time.ticks_ms()
    if time.ticks_diff(t, last_file_write) > 50:
        write_log("tick")
        last_file_write = t


def irq(event, data):
    write_log("event", event)

    if event == _IRQ_CENTRAL_CONNECT:
        print("_IRQ_CENTRAL_CONNECT")
        waiting_events[event] = data[0]
    elif event == _IRQ_CENTRAL_DISCONNECT:
        print("_IRQ_CENTRAL_DISCONNECT")
    elif event == _IRQ_PERIPHERAL_CONNECT:
        print("_IRQ_PERIPHERAL_CONNECT")
        waiting_events[event] = data[0]
    elif event == _IRQ_PERIPHERAL_DISCONNECT:
        print("_IRQ_PERIPHERAL_DISCONNECT")
    elif event == _IRQ_GATTC_SERVICE_RESULT:
        # conn_handle, start_handle, end_handle, uuid = data
        if data[-1] == SERVICE_UUID:
            print("_IRQ_GATTC_SERVICE_RESULT", data[3])
            waiting_events[event] = (data[1], data[2])
        else:
            return
    elif event == _IRQ_GATTC_SERVICE_DONE:
        print("_IRQ_GATTC_SERVICE_DONE")
    elif event == _IRQ_GATTC_CHARACTERISTIC_RESULT:
        # conn_handle, def_handle, value_handle, properties, uuid = data
        if data[-1] == CHAR_UUID:
            print("_IRQ_GATTC_CHARACTERISTIC_RESULT", data[-1])
            waiting_events[event] = data[2]
        else:
            return
    elif event == _IRQ_GATTC_CHARACTERISTIC_DONE:
        print("_IRQ_GATTC_CHARACTERISTIC_DONE")
    elif event == _IRQ_GATTC_READ_RESULT:
        print("_IRQ_GATTC_READ_RESULT", bytes(data[-1]))
    elif event == _IRQ_GATTC_READ_DONE:
        print("_IRQ_GATTC_READ_DONE", data[-1])
    elif event == _IRQ_GATTC_WRITE_DONE:
        print("_IRQ_GATTC_WRITE_DONE", data[-1])
    elif event == _IRQ_GATTS_WRITE:
        print("_IRQ_GATTS_WRITE")
    elif event == _IRQ_GATTS_READ_REQUEST:
        print("_IRQ_GATTS_READ_REQUEST")

    if event not in waiting_events:
        waiting_events[event] = None


def wait_for_event(event, timeout_ms):
    t0 = time.ticks_ms()
    while time.ticks_diff(time.ticks_ms(), t0) < timeout_ms:
        periodic_log_write()
        if event in waiting_events:
            return waiting_events.pop(event)
        machine.idle()
    raise ValueError("Timeout waiting for {}".format(event))


# Acting in peripheral role.
def instance0():
    global log_file
    log_file = open(LOG_PATH_INSTANCE0, "w")
    write_log("start")
    ble.active(1)
    ble.irq(irq)
    multitest.globals(BDADDR=ble.config("mac"))
    ((char_handle,),) = ble.gatts_register_services(SERVICES)
    multitest.next()
    try:
        for repeat in range(2):
            print("gap_advertise")
            ble.gap_advertise(50_000, b"\x02\x01\x06\x04\xffMPY")
            # Wait for central to connect, do a sequence of read/write, then disconnect.
            wait_for_event(_IRQ_CENTRAL_CONNECT, TIMEOUT_MS)
            for op in range(4):
                wait_for_event(_IRQ_GATTS_READ_REQUEST, TIMEOUT_MS)
                wait_for_event(_IRQ_GATTS_WRITE, TIMEOUT_MS)
            wait_for_event(_IRQ_CENTRAL_DISCONNECT, 2 * TIMEOUT_MS)
    finally:
        ble.active(0)
        log_file.close()
        os.unlink(LOG_PATH_INSTANCE0)


# Acting in central role.
def instance1():
    global log_file
    log_file = open(LOG_PATH_INSTANCE1, "w")
    write_log("start")
    ble.active(1)
    ble.irq(irq)
    multitest.next()
    try:
        for repeat in range(2):
            # Connect to peripheral and then disconnect.
            print("gap_connect")
            ble.gap_connect(BDADDR[0], BDADDR[1], 5000)
            conn_handle = wait_for_event(_IRQ_PERIPHERAL_CONNECT, TIMEOUT_MS)

            # Discover services.
            print("gattc_discover_services")
            ble.gattc_discover_services(conn_handle)
            start_handle, end_handle = wait_for_event(_IRQ_GATTC_SERVICE_RESULT, TIMEOUT_MS)
            wait_for_event(_IRQ_GATTC_SERVICE_DONE, TIMEOUT_MS)

            # Discover characteristics.
            print("gattc_discover_characteristics")
            ble.gattc_discover_characteristics(conn_handle, start_handle, end_handle)
            value_handle = wait_for_event(_IRQ_GATTC_CHARACTERISTIC_RESULT, TIMEOUT_MS)
            wait_for_event(_IRQ_GATTC_CHARACTERISTIC_DONE, TIMEOUT_MS)

            for op in range(4):
                print("gattc_read")
                ble.gattc_read(conn_handle, value_handle)
                wait_for_event(_IRQ_GATTC_READ_RESULT, TIMEOUT_MS)
                wait_for_event(_IRQ_GATTC_READ_DONE, TIMEOUT_MS)
                print("gattc_write")
                ble.gattc_write(conn_handle, value_handle, "{}".format(op), 1)
                wait_for_event(_IRQ_GATTC_WRITE_DONE, TIMEOUT_MS)

            # Disconnect.
            print("gap_disconnect:", ble.gap_disconnect(conn_handle))
            wait_for_event(_IRQ_PERIPHERAL_DISCONNECT, 2 * TIMEOUT_MS)
    finally:
        ble.active(0)
        log_file.close()
        os.unlink(LOG_PATH_INSTANCE1)


ble = bluetooth.BLE()
